from functools import wraps
from io import StringIO
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union

from lsprotocol.types import (CompletionItem, CompletionList,
                              DocumentHighlight, DocumentSymbol, Hover,
                              Position, Range, SemanticTokens,
                              SymbolInformation, TextDocumentSaveReason,
                              TextEdit)

from llanguage_test_client.lsp_client import LspClient


def cantor_pair(k1: int, k2: int) -> int:
    return (((k1 + k2) * (k1 + k2 + 1)) // 2) + k2


def cantor_tuple(*ks: int) -> int:
    match len(ks):
        case 0:
            return 0
        case 1:
            k1 = ks[0]
            return k1
        case 2:
            k1, k2 = ks
            return cantor_pair(k1, k2)
        case _:
            kn = ks[-1]
            return cantor_pair(cantor_tuple(*ks[:-1]), kn)


def edit_key_fn(edit: TextEdit) -> int:
    """
    Orders TextEdit elements order according to their starting and ending
    positions.
    """
    range = edit.range
    start = range.start
    end = range.end
    return cantor_tuple(
        start.line, start.character,
        end.line, end.character
    )


def requires_path(fn: Callable) -> Callable:
    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        if self._path is not None:
            fn(self, *args, **kwargs)
    return wrapper


ChangeListener = Callable[[int, Optional[str]], None]


class LspTextDocument:
    client: LspClient
    document_id: int
    language_id: str
    _path: Optional[Path]

    version: int
    buf: StringIO
    is_new: bool
    pos_by_line: List[int]
    len_by_line: List[int]

    change_listeners: List[ChangeListener]

    selection: Optional[Range]
    symbol_highlights: Optional[List[DocumentHighlight]]
    preview: Optional[Hover]
    symbols: Union[List[SymbolInformation], List[DocumentSymbol], None]
    semantic_highlights: Union[SemanticTokens, None]
    completions: Union[List[CompletionItem], CompletionList, None]

    def __init__(
            self,
            client: LspClient,
            document_id: int,
            language_id: str,
            path: Optional[Union[Path, str]] = None
    ) -> None:
        self.client = client
        self.document_id = document_id
        self.language_id = language_id
        if path is not None:
            self.path = path
        else:
            self._path = None
        self.version = 0
        self.buf = StringIO()
        self.pos_by_line = [0]
        self.len_by_line = [1]
        self.is_new = (path is None)
        self.selection = None
        self.symbol_highlights = None
        self.preview = None
        self.symbols = None
        self.change_listeners = []
        self.semantic_highlights = None
        self.completions = None

    def on_change(self, listener: ChangeListener) -> None:
        self.change_listeners.append(listener)

    def signal_change(self) -> None:
        uri = self.uri if self._path is not None else None
        for change_fn in self.change_listeners:
            change_fn(self.document_id, uri)

    @property
    def path(self) -> Path:
        if self._path is not None:
            return self._path
        raise RuntimeError('`path` must be set before referencing it.')

    @path.setter
    def path(self, path: Optional[Union[Path, str]]) -> None:
        match path:
            case str():
                self._path = Path(path)
            case Path():
                self._path = path
            case None:
                pass
        if self._path is not None:
            if not self._path.exists():
                self._path.touch()
            self._path = self._path.absolute()

    @property
    def uri(self) -> str:
        return f"file://{self.path}"

    @property
    def text(self) -> str:
        return self.buf.getvalue()

    def __len__(self) -> int:
        return len(self.text)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, LspTextDocument) \
            and self.document_id == other.document_id

    def __hash__(self) -> int:
        return hash((self.document_id,))

    @property
    def cursor(self) -> Tuple[int, int]:
        line, column = self.pos_to_linecol(self.position)
        return line + 1, column + 1

    @cursor.setter
    def cursor(self, line_and_col: Tuple[int, int]) -> None:
        line, col = line_and_col
        if line < 0:
            line = len(self.pos_by_line) + line
        else:
            line -= 1
        if col < 0:
            col = self.len_by_line[line] + col
        else:
            col -= 1
        self.seek(line, col)

    @property
    def position(self) -> int:
        return self.buf.tell()

    @position.setter
    def position(self, position: int) -> None:
        num_chars = len(self.text)
        if position < num_chars:
            self.buf.seek(position)
        else:
            raise ValueError(
                f'Position out-of-bounds for text of length {num_chars}: {position}'
            )

    @property
    def line(self) -> int:
        line, _ = self.cursor
        return line

    @property
    def column(self) -> int:
        _, column = self.cursor
        return column

    def recompute_indices(self) -> None:
        self.pos_by_line.clear()
        self.len_by_line.clear()
        col = 0
        self.pos_by_line.append(0)
        for pos, char in enumerate(self.text):
            if char == '\n':
                self.pos_by_line.append(pos + 1)
                self.len_by_line.append(col + 1)
                col = 0
            else:
                col += 1
        self.len_by_line.append(col + 1)

    def load(self) -> None:
        if self._path is not None:
            with open(self.path, "r", encoding="utf-8") as f:
                self.buf.write(f.read())
            self.recompute_indices()
            self.client.text_document_did_open(
                self.document_id,
                self.uri,
                self.language_id,
                self.bump_version(),
                self.text
            )
            self.signal_change()
        else:
            raise RuntimeError('`path` must be set before calling `load()`')

    def close(self) -> None:
        uri = self.uri if self._path is not None else None
        self.client.text_document_did_close(self.document_id, uri)

    def save(self, path: Optional[Union[Path, str]] = None) -> None:
        self.path = path
        if self._path is None:
            raise RuntimeError('`path` must be set before calling `save()`')
        if not self.is_new:
            reason = TextDocumentSaveReason.Manual
            self.client.text_document_will_save_wait_until(self.uri, reason)
            self.client.text_document_will_save(self.uri, reason)
        else:
            self.client.workspace_will_create_files([self.uri])
        with open(self.path, "w+", encoding="utf-8") as f:
            f.write(self.text)
        if self.is_new:
            self.client.workspace_did_create_files([self.uri])
            self.client.text_document_did_open(
                self.document_id,
                self.uri,
                self.language_id,
                self.bump_version(),
                self.text
            )
            self.is_new = False
        self.client.text_document_did_save(self.uri, self.text)

    def move(self, path: Union[Path, str]) -> None:
        old_path = self._path
        if old_path is not None:
            old_uri = self.uri
            self.path = path
            new_path = self.path
            if old_path != new_path:
                new_uri = self.uri
                files = [
                    (old_uri, new_uri),
                ]
                self.client.workspace_will_rename_files(files)
                old_path.replace(new_path)
                self.client.workspace_did_rename_files(files)
        else:
            raise RuntimeError('Must call `save()` before `rename()`')

    def remove(self) -> None:
        if self._path is None:
            raise RuntimeError('`path` must be set before calling `remove()`')
        uri = self.uri
        self.client.workspace_will_delete_files([uri])
        self._path.unlink()
        self._path = None
        self.client.workspace_did_delete_files([uri])

    def bump_version(self) -> int:
        self.version += 1
        return self.version

    def pos_to_linecol(self, position: int) -> Tuple[int, int]:
        num_chars = len(self.text)
        if position <= num_chars:
            num_lines = len(self.len_by_line)
            lower = 0
            upper = num_lines - 1
            line = (lower + upper + 1) // 2
            column = position - self.pos_by_line[line]
            while (column < 0) or (column >= self.len_by_line[line]):
                if column < 0:
                    upper = line - 1
                else:
                    lower = line
                line = (lower + upper + 1) // 2
                column = position - self.pos_by_line[line]
            return line, column
        raise ValueError(
            f'Position out-of-bounds for text of length {num_chars}: {position}'
        )

    def linecol_to_pos(self, line: int, column: int) -> int:
        num_lines = len(self.pos_by_line)
        if line < num_lines:
            num_columns = self.len_by_line[line]
            if column <= num_columns:
                position = self.pos_by_line[line] + column
                return position
            raise ValueError(
                f'Column out-of-bounds for line {line} with {num_columns} column(s): {column}'
            )
        raise ValueError(
            f'Line out-of-bounds for {num_lines} line(s): {line}'
        )

    def seek(self, line: int, column: int) -> None:
        """
        Moves the cursor to the given location (0-indexed line and column).
        """
        position = self.linecol_to_pos(line, column)
        self.buf.seek(position)

    def select(
            self,
            start_line: int,
            start_column: int,
            end_line: int,
            end_column: int
    ) -> None:
        start = Position(start_line - 1, start_column - 1)
        end = Position(end_line - 1, end_column - 1)
        self.selection = Range(start, end)
        if (start_line < end_line) \
           or ((start_line == end_line) and (start_column <= end_column)):
            self.seek(end_line - 1, end_column - 1)
        else:
            self.seek(start_line - 1, start_column - 1)

    def delete(self, length: int = 1) -> None:
        """
        Deletes characters from the current location.
        """
        start = self.buf.tell()
        start_line, start_column = self.pos_to_linecol(start)
        end = min(start + length, len(self.text))
        end_line, end_column = self.pos_to_linecol(end)
        self.buf.seek(end)
        suffix = self.buf.read()
        self.buf.seek(start)
        self.buf.write(suffix)
        self.buf.truncate()
        self.buf.seek(start)
        if self._path is not None:
            self.client.text_document_did_change(
                self.uri, self.bump_version(),
                start_line, start_column,
                end_line, end_column,
                self.text, ""
            )
        self.recompute_indices()
        self.signal_change()

    def backspace(self, length: int = 1) -> None:
        """
        Deletes characters up to the current location.
        """
        stop = self.buf.tell()
        start = stop - length
        self.buf.seek(start)
        self.delete(length)

    def write(self, text: str) -> None:
        """
        Inserts text at the current location (does not overwrite existing text).
        """
        start = self.position
        end = start
        start_line, start_column = self.pos_to_linecol(start)
        end_line, end_column = self.pos_to_linecol(end)
        suffix = self.buf.read()
        self.buf.seek(start)
        self.buf.write(text)
        self.buf.write(suffix)
        if self._path is not None:
            self.client.text_document_did_change(
                self.uri, self.bump_version(),
                start_line, start_column,
                end_line, end_column,
                self.text, text
            )
        self.recompute_indices()
        self.signal_change()

    def newline(self) -> None:
        """
        Inserts a carriage return at the current location.
        """
        self.write('\n')

    def replace(self, text: str) -> None:
        """
        Overwrites (substitutes) text at the current location.
        """
        start = self.position
        end = min(start + len(text), len(self.text))
        start_line, start_column = self.pos_to_linecol(start)
        end_line, end_column = self.pos_to_linecol(end)
        self.buf.write(text)
        if self._path is not None:
            self.client.text_document_did_change(
                self.uri, self.bump_version(),
                start_line, start_column,
                end_line, end_column,
                self.text, text
            )
        self.recompute_indices()
        self.signal_change()

    def apply(self, edits: List[TextEdit]) -> None:
        origin = self.position
        for edit in sorted(edits, key=edit_key_fn, reverse=True):
            range = edit.range
            start = range.start
            end = range.end
            self.seek(end.line, end.character)
            suffix = self.buf.read()
            self.seek(start.line, start.character)
            self.buf.write(edit.new_text)
            # Append the newline if it was replaced
            if end.character == self.len_by_line[end.line]:
                self.buf.write('\n')
            self.buf.write(suffix)
            self.buf.truncate()
        origin = min(origin, len(self.text) - 1)
        self.buf.seek(origin)
        self.recompute_indices()

    @requires_path
    def goto_definition(self) -> None:
        line, column = self.pos_to_linecol(self.position)
        self.client.goto_definition(self.uri, line, column)

    @requires_path
    def rename(self, new_name: str) -> None:
        line, column = self.pos_to_linecol(self.position)
        self.client.rename(self.uri, line, column, new_name)

    @requires_path
    def highlight_symbol(self) -> None:
        line, column = self.pos_to_linecol(self.position)
        self.client.highlight_symbol(self.uri, line, column)

    @requires_path
    def hover(self) -> None:
        line, column = self.pos_to_linecol(self.position)
        self.client.hover(self.uri, line, column)

    @requires_path
    def semantic_highlight(self) -> None:
        self.client.semantic_highlight(self.uri)

    @requires_path
    def complete(self) -> None:
        line, column = self.pos_to_linecol(self.position)
        self.client.complete(self.uri, line, column)

    @requires_path
    def format(self) -> None:
        self.client.format(self.uri)

    @requires_path
    def format_range(self) -> None:
        if self.selection is not None:
            self.client.format_range(self.uri, self.selection)



#!/usr/bin/env python3

"""
This script generates the parser.yy.patch that makes the GLR parser a LALR(1)
parser.

Usage:

    ci/generate_lalr1_patch.py

"""

import subprocess

# Check if git diff shows any changes using return code
result = subprocess.run(['git', 'diff'], capture_output=True, text=True)
if result.returncode != 0:
    print("Git diff shows changes. Please commit the changes first.")
    exit(1)

# Define file paths
file_path = 'src/lfortran/parser/parser.yy'
patch_path = 'ci/parser.yy.patch'

# Read the file
with open(file_path, 'r') as f:
    lines = f.readlines()

# Modify id section
for j in range(len(lines) - 1):
    if lines[j].strip() == 'id' and \
       ': TK_NAME { $$ = SYMBOL($1, @$); }' in lines[j+1]:
        # Delete all subsequent | KW_* lines until ;
        k = j + 2
        while k < len(lines) and lines[k].strip() != ';':
            stripped_line = lines[k].strip()
            if not stripped_line.startswith('| KW'):
                raise ValueError(f"Unexpected line in id section: {stripped_line}")
            del lines[k]
            # No k += 1, as del shifts the list
        break

# Write modified lines back to the file
with open(file_path, 'w') as f:
    f.writelines(lines)

# Generate the patch using git diff
with open(patch_path, 'w') as patch_file:
    subprocess.run(['git', 'diff', '--', file_path], stdout=patch_file, check=True)

# Restore the original file to remove changes
subprocess.run(['git', 'restore', file_path], check=True)

print(f"Patch generated and saved to {patch_path}")



#!/usr/bin/env python

import os
import subprocess as sp
import sys
from typing import Dict

ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__)))
sys.path.append(os.path.join(ROOT_DIR, "src", "libasr"))

from compiler_tester.tester import color, fg, log, run_test, style, tester_main

def run_cmd(cmd, cwd=None):
    print(f"+ {cmd}")
    process = sp.run(cmd, shell=True, cwd=cwd)
    if process.returncode != 0:
        print("Command failed.")
        exit(1)

def single_test(test: Dict, verbose: bool, no_llvm: bool, skip_run_with_dbg: bool,
                update_reference: bool, verify_hash: bool,
                no_color: bool, specific_backends=None,
                excluded_backends=None) -> None:
    def is_included(backend):
        return test.get(backend, False) \
            and (specific_backends is None or backend in specific_backends) \
            and (excluded_backends is None or backend not in excluded_backends)

    filename = test["filename"]
    show_verbose = "" if not verbose else "-v"
    tokens = is_included("tokens")
    ast = is_included("ast")
    ast_indent = is_included("ast_indent")
    ast_disable_style_suggestion = is_included("ast_disable_style_suggestion")
    ast_json = is_included("ast_json")
    ast_no_prescan = is_included("ast_no_prescan")
    ast_f90 = is_included("ast_f90")
    ast_cpp = is_included("ast_cpp")
    ast_cpp_hip = is_included("ast_cpp_hip")
    ast_openmp = is_included("ast_openmp")
    lookup_name = is_included("lookup_name")
    rename_symbol = is_included("rename_symbol")
    line = "-1"
    if is_included("line"):
        line = str(test["line"])
    column = "-1"
    if is_included("column"):
        column = str(test["column"])
    asr = is_included("asr")
    asr_ignore_pragma = is_included("asr_ignore_pragma")
    asr_implicit_typing = is_included("asr_implicit_typing")
    asr_disable_implicit_typing = is_included("asr_disable_implicit_typing")
    enable_and_disable_implicit_typing = is_included("enable_and_disable_implicit_typing")
    asr_implicit_interface = is_included("asr_implicit_interface")
    asr_implicit_interface_and_typing = is_included("asr_implicit_interface_and_typing")
    asr_implicit_argument_casting = is_included("asr_implicit_argument_casting")
    enable_disable_implicit_argument_casting = is_included("enable_disable_implicit_argument_casting")
    asr_implicit_interface_and_typing_with_llvm = is_included("asr_implicit_interface_and_typing_with_llvm")
    asr_disable_warnings = is_included("asr_disable_warnings")
    asr_disable_style_suggestion_and_warnings = is_included("asr_disable_style_suggestion_and_warnings")
    asr_enable_style_suggestion = is_included("asr_enable_style_suggestion")
    enable_disable_style_suggestion = is_included("enable_disable_style_suggestion")
    continue_compilation = is_included("continue_compilation")
    fixed_form_cc_asr = is_included("fixed_form_cc_asr")
    semantics_only_cc = is_included("semantics_only_cc")
    show_errors = is_included("show_errors")
    document_symbols = is_included("document_symbols")
    syntax_only_cc = is_included("syntax_only_cc")
    show_asr_with_cc = is_included("show_asr_with_cc")
    asr_use_loop_variable_after_loop = is_included("asr_use_loop_variable_after_loop")
    asr_preprocess = is_included("asr_preprocess")
    asr_indent = is_included("asr_indent")
    asr_json = is_included("asr_json")
    asr_clojure = is_included("asr_clojure")
    asr_openmp = is_included("asr_openmp")
    c_target_omp = is_included("c_target_omp")
    c_target_cuda = is_included("c_target_cuda")
    asr_logical_casting = is_included("asr_logical_casting")
    mod_to_asr = is_included("mod_to_asr")
    llvm = is_included("llvm")
    cpp = is_included("cpp")
    cpp_infer = is_included("cpp_infer")
    c = is_included("c")
    is_cumulative_pass = is_included("cumulative")
    julia = is_included("julia")
    wat = is_included("wat")
    obj = is_included("obj")
    x86 = is_included("x86")
    fortran = is_included("fortran")
    bin_ = is_included("bin")
    fast = is_included("fast")
    print_leading_space = is_included("print_leading_space")
    interactive = is_included("interactive")
    options = test.get("options", "")
    pass_ = test.get("pass", None)
    extrafiles = test.get("extrafiles", "").split(",")
    run = test.get("run")
    run_with_dbg = test.get("run_with_dbg")
    optimization_passes = ["flip_sign", "div_to_mul", "fma", "sign_from_value",
                           "inline_function_calls", "loop_unroll",
                           "dead_code_removal"]

    if pass_ is not None:
        pass_list = pass_.split(",")

        for _pass in pass_list:
            _pass = _pass.rstrip(" ").lstrip(" ")
            if (_pass not in ["do_loops", "global_stmts",
                        "transform_optional_argument_functions",
                        "array_op", "select_case",
                        "class_constructor", "implied_do_loops",
                        "pass_array_by_data", "init_expr", "where",
                        "nested_vars", "insert_deallocate", "openmp",
                        "array_struct_temporary"] and
                _pass not in optimization_passes):
                raise Exception(f"Unknown pass: {_pass}")
    if update_reference:
        log.debug(f"{color(style.bold)} UPDATE TEST: {color(style.reset)} {filename}")
    elif verify_hash:
        log.debug(f"{color(style.bold)} VERIFY HASH: {color(style.reset)} {filename}")
    else:
        log.debug(f"{color(style.bold)} START TEST: {color(style.reset)} {filename}")

    extra_args = f"--no-error-banner {show_verbose}"
    if print_leading_space:
        extra_args += " --print-leading-space"
    if interactive:
        extra_args += " --interactive-parse"
    if cpp_infer:
        extra_args += " --cpp-infer"
    if line:
        extra_args += " --line=" + line
    if column:
        extra_args += " --column=" + column
    if options:
        extra_args += " " + options

    if tokens:
        run_test(
            filename,
            "tokens",
            "lfortran --no-color --show-tokens {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)
    if ast:
        if filename.endswith(".f"):
            # Use fixed form
            run_test(
                filename,
                "ast",
                "lfortran --fixed-form --show-ast --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
        else:
            run_test(
                filename,
                "ast",
                "lfortran --show-ast --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
    if ast_indent:
        run_test(
            filename,
            "ast_indent",
            "lfortran --show-ast --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if ast_openmp:
        run_test(
            filename,
            "ast_openmp",
            "lfortran --show-ast --no-color --openmp {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if enable_and_disable_implicit_typing:
        if no_llvm:
            log.info(f"{filename} * obj    SKIPPED as requested")
        else:
            run_test(filename, "run", "lfortran --implicit-typing --disable-implicit-typing --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if enable_disable_implicit_argument_casting:
        if no_llvm:
            log.info(f"{filename} * obj    SKIPPED as requested")
        else:
            run_test(filename, "run", "lfortran --implicit-argument-casting --disable-implicit-argument-casting --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if ast_json:
        run_test(
            filename,
            "ast_json",
            "lfortran --show-ast --no-indent --json {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if ast_disable_style_suggestion:
        run_test(
            filename,
            "ast_disable_style_suggestion",
            "lfortran --show-ast --no-style-suggestions --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if ast_no_prescan:
        # Use free form with prescan disabled
        run_test(
            filename,
            "ast_no_prescan",
            "lfortran --no-prescan --show-ast --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if ast_f90:
        if filename.endswith(".f"):
            # Use fixed form
            run_test(
                filename,
                "ast_f90",
                "lfortran --fixed-form --show-ast-f90 --no-indent --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
        else:
            # Use free form
            run_test(
                filename,
                "ast_f90",
                "lfortran --show-ast-f90 --no-indent --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if lookup_name:
        run_test(
            filename,
            "lookup_name",
            "lfortran --lookup-name --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)
    if rename_symbol:
        run_test(
            filename,
            "rename_symbol",
            "lfortran --rename-symbol --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_disable_warnings:
        run_test(
            filename,
            "asr_disable_warnings",
            "lfortran --show-asr --no-warnings --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_disable_style_suggestion_and_warnings:
        run_test(
            filename,
            "asr_disable_style_suggestion_and_warnings",
            "lfortran --show-asr --no-style-suggestions --no-warnings --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)
    
    if asr_enable_style_suggestion:
        if no_llvm:
            log.info(f"{filename} * obj    SKIPPED as requested")
        else:
            run_test(
                filename,
                "asr_enable_style_suggestion",
                "lfortran --std=f23 --style-suggestions --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
    
    if enable_disable_style_suggestion:
        if no_llvm:
            log.info(f"{filename} * obj    SKIPPED as requested")
        else:
            run_test(
                filename,
                "enable_disable_style_suggestion",
                "lfortran --style-suggestions --no-style-suggestions --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if asr:
        # run fixed form
        if filename.endswith(".f"):
            run_test(
                filename,
                "asr",
                "lfortran --fixed-form --show-asr --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
        else:
            skip_test = False
            for extrafile in extrafiles:
                extrafile_ = extrafile.rstrip().lstrip()

                if no_llvm and len(extrafile_) > 0:
                    log.info(f"{filename} * asr   SKIPPED because LLVM is not enabled")
                    skip_test = True
                    break

                if len(extrafile_) > 0:
                    extrafile_ = os.path.join("tests", extrafile_)
                    modfile = extrafile_[:-4] + ".mod"
                    if not os.path.exists(modfile):
                        run_cmd("lfortran -c {}".format(extrafile_))

            if not skip_test:
                run_test(
                    filename,
                    "asr",
                    "lfortran --show-asr --no-color {infile} -o {outfile}",
                    filename,
                    update_reference,
                    verify_hash,
                    extra_args)

                if pass_ is not None:
                    if fast:
                        cmd = "lfortran --pass=" + pass_ + \
                            " --show-asr --no-color --fast {infile} -o {outfile}"
                    else:
                        cmd = "lfortran --pass=" + pass_ + \
                            " --show-asr --no-color {infile} -o {outfile}"
                    pass_ = pass_.replace(",", "_")
                    run_test(filename, "pass_{}".format(pass_), cmd,
                            filename,
                            update_reference,
                            verify_hash,
                            extra_args)

            pass_ = None

    if asr_implicit_interface_and_typing:
        # run fixed form
        if filename.endswith(".f"):
            run_test(
                filename,
                "asr",
                "lfortran --fixed-form --show-asr --implicit-typing --implicit-interface --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
        else:
            run_test(
                filename,
                "asr",
                "lfortran --show-asr --implicit-typing --implicit-interface --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
    if asr_use_loop_variable_after_loop:
        run_test(
            filename,
            "asr",
            "lfortran --show-asr --use-loop-variable-after-loop --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_implicit_argument_casting:
        run_test(
            filename,
            "asr",
            "lfortran --show-asr --implicit-argument-casting --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_implicit_interface_and_typing_with_llvm:
        if no_llvm:
            log.info(f"{filename} * llvm   SKIPPED as requested")
        else:
            run_test(
                filename,
                "llvm",
                "lfortran --show-llvm --implicit-typing --implicit-interface {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if semantics_only_cc:
        run_test(filename, "asr", "lfortran --semantics-only --continue-compilation --no-color {infile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if document_symbols:
        skip_test = False
        for extrafile in extrafiles:
            extrafile_ = extrafile.rstrip().lstrip()

            if no_llvm and len(extrafile_) > 0:
                log.info(f"{filename} * asr   SKIPPED because LLVM is not enabled")
                skip_test = True
                break

            if len(extrafile_) > 0:
                extrafile_ = os.path.join("tests", extrafile_)
                modfile = extrafile_[:-4] + ".mod"
                if not os.path.exists(modfile):
                    run_cmd("lfortran -c {}".format(extrafile_))
        if not skip_test:
            run_test(filename, "asr", "lfortran --show-document-symbols --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if show_errors:
        run_test(filename, "asr", "lfortran --show-errors --continue-compilation --no-color {infile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if syntax_only_cc:
        run_test(filename, "ast", "lfortran --continue-compilation --show-ast --no-color {infile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if show_asr_with_cc:
        run_test(filename, "asr", "lfortran --continue-compilation --show-asr --no-color {infile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if continue_compilation:
        if no_llvm:
            log.info(f"{filename} * obj    SKIPPED as requested")
        else:
            run_test(filename, "run", "lfortran --continue-compilation --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if fixed_form_cc_asr:
        run_test(
                filename,
                "asr",
                "lfortran --fixed-form --continue-compilation --show-asr --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if asr_implicit_typing:
        run_test(
            filename,
            "asr",
            "lfortran --show-asr --implicit-typing --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)
    
    if asr_disable_implicit_typing:
        if no_llvm:
            log.info(f"{filename} * llvm   SKIPPED as requested")
        run_test(
            filename,
            "asr",
            "lfortran --std=f23 --show-asr --disable-implicit-typing --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_implicit_interface:
        if filename.endswith(".f"):
            run_test(
                filename,
                "asr",
                "lfortran --fixed-form --implicit-interface --show-asr --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)
        else:
            run_test(
                filename,
                "asr",
                "lfortran --show-asr --implicit-interface --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if asr_preprocess:
        run_test(
            filename,
            "asr_preprocess",
            "lfortran --cpp --show-asr --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_indent:
        run_test(
            filename,
            "asr_indent",
            "lfortran --show-asr --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_json:
        run_test(
            filename,
            "asr_json",
            "lfortran --show-asr --no-indent --json {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_clojure:
        run_test(
            filename,
            "asr_clojure",
            "lfortran --show-asr --no-color --no-indent --clojure {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_openmp:
        run_test(
            filename,
            "asr_openmp",
            "lfortran --show-asr --no-color --openmp {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)
        
    if c_target_omp:
        run_test(
            filename,
            "c_target_omp",
            "lfortran --show-c --openmp {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)
        
    if c_target_cuda:
        run_test(
            filename,
            "c_target_cuda",
            "lfortran --show-c --openmp --target-offload {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if asr_logical_casting:
        run_test(
            filename,
            "asr_logical_casting",
            "lfortran --show-asr --logical-casting --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if mod_to_asr:
        run_test(
            filename,
            "mod_to_asr",
            "lfortran mod --show-asr --no-indent --no-color {infile}",
            filename,
            update_reference,
            verify_hash)

    if asr_ignore_pragma:
        run_test(
            filename,
            "asr_ignore_pragma",
            "lfortran --ignore-pragma --show-asr --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if pass_ is not None:
        cmd = "lfortran "
        if is_cumulative_pass:
            cmd += "--cumulative "
        cmd += "--pass=" + pass_ + \
            " --show-asr --no-color {infile} -o {outfile}"
        pass_ = pass_.replace(",", "_")
        run_test(filename, "pass_{}".format(pass_), cmd,
                filename,
                update_reference,
                verify_hash,
                extra_args)
    if llvm:
        if no_llvm:
            log.info(f"{filename} * llvm   SKIPPED as requested")
        else:
            run_test(
                filename,
                "llvm",
                "lfortran --no-color --show-llvm {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if cpp:
        run_test(filename, "cpp", "lfortran --no-color --show-cpp {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if obj:
        if no_llvm:
            log.info(f"{filename} * obj    SKIPPED as requested")
        else:
            run_test(
                filename,
                "obj",
                "lfortran --no-color -c {infile} -o output.o",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if c:
        run_test(filename, "c", "lfortran --no-color --show-c {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if julia:
        run_test(filename, "julia", "lfortran --no-color --show-julia {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if wat:
        run_test(filename, "wat", "lfortran --no-color --show-wat {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if x86:
        run_test(
            filename,
            "x86",
            "lfortran --no-color --backend=x86 {infile} -o output",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if fortran:
        run_test(
            filename,
            "fortran",
            "lfortran --show-fortran --no-color {infile} -o {outfile}",
            filename,
            update_reference,
            verify_hash,
            extra_args)

    if bin_:
        run_test(filename, "bin", "lfortran --no-color {infile} -o {outfile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if run:
        if no_llvm:
            log.info(f"{filename} * obj    SKIPPED as requested")
        else:
            run_test(filename, "run", "lfortran --no-color {infile}",
                filename,
                update_reference,
                verify_hash,
                extra_args)

    if run_with_dbg:
        if skip_run_with_dbg:
            log.info(f"{filename} * run_with_dbg   SKIPPED as requested")
        else:
            run_test(
                filename, "run_dbg",
                "lfortran {infile} -g --no-color",
            filename,
            update_reference,
            verify_hash,
            extra_args)

if __name__ == "__main__":
    tester_main("LFortran", single_test)



import json
import os
import re
import signal
import subprocess
import sys
import threading
import time
import traceback
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import wraps
from io import BytesIO
from pathlib import Path
from typing import (IO, Any, BinaryIO, Callable, Dict, Iterator, List,
                    Optional, Tuple, Union)

import psutil

from cattrs import Converter

from lsprotocol import converters
from lsprotocol.types import (ClientCapabilities,
                              ClientRegisterCapabilityRequest,
                              ClientRegisterCapabilityResponse,
                              CompletionParams, CreateFilesParams,
                              DeleteFilesParams,
                              DidChangeConfigurationClientCapabilities,
                              DidChangeConfigurationParams,
                              DidChangeTextDocumentParams,
                              DidCloseTextDocumentParams,
                              DidOpenTextDocumentParams,
                              DidSaveTextDocumentParams,
                              DocumentFormattingParams,
                              DocumentHighlightParams,
                              DocumentRangeFormattingParams,
                              DocumentSymbolParams, ExitNotification,
                              FileCreate, FileDelete,
                              FileOperationClientCapabilities, FileRename,
                              FormattingOptions, HoverParams,
                              InitializedNotification, InitializedParams,
                              InitializeParams, InitializeRequest,
                              InitializeResponse,
                              InitializeResultServerInfoType, Position, Range,
                              Registration, RenameFilesParams, SaveOptions,
                              SemanticTokensParams, ServerCapabilities,
                              ShutdownRequest, TelemetryEventNotification,
                              TextDocumentClientCapabilities,
                              TextDocumentCompletionRequest,
                              TextDocumentContentChangeEvent_Type1,
                              TextDocumentContentChangeEvent_Type2,
                              TextDocumentDidChangeNotification,
                              TextDocumentDidCloseNotification,
                              TextDocumentDidOpenNotification,
                              TextDocumentDidSaveNotification,
                              TextDocumentDocumentHighlightRequest,
                              TextDocumentDocumentSymbolRequest,
                              TextDocumentFormattingRequest,
                              TextDocumentHoverRequest, TextDocumentIdentifier,
                              TextDocumentItem,
                              TextDocumentRangeFormattingRequest,
                              TextDocumentSaveReason,
                              TextDocumentSemanticTokensFullRequest,
                              TextDocumentSyncClientCapabilities,
                              TextDocumentSyncKind, TextDocumentSyncOptions,
                              TextDocumentWillSaveNotification,
                              TextDocumentWillSaveWaitUntilRequest,
                              TextDocumentWillSaveWaitUntilResponse,
                              VersionedTextDocumentIdentifier,
                              WillSaveTextDocumentParams,
                              WorkspaceClientCapabilities,
                              WorkspaceConfigurationRequest,
                              WorkspaceConfigurationResponse,
                              WorkspaceDidChangeConfigurationNotification,
                              WorkspaceDidCreateFilesNotification,
                              WorkspaceDidDeleteFilesNotification,
                              WorkspaceDidRenameFilesNotification,
                              WorkspaceWillCreateFilesRequest,
                              WorkspaceWillDeleteFilesRequest,
                              WorkspaceWillRenameFilesRequest)

from llanguage_test_client.json_rpc import JsonObject, JsonValue
from llanguage_test_client.lsp_client import FileRenameMapping, LspClient, Uri
from llanguage_test_client.lsp_json_stream import LspJsonStream
from llanguage_test_client.lsp_text_document import LspTextDocument


def requires_server_capabilities(fn: Callable) -> Callable:
    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        # NOTE: If an error occurred during the server's initialization, the
        # `server_capabilities` attribute may never have been set.
        if hasattr(self, 'server_capabilities'):
            return fn(self, *args, **kwargs)
        return False
    return wrapper


# NOTE: File URIs follow one of the following schemes:
# 1. `file:/path` (no hostname)
# 2. `file:///path` (empty hostname)
# 3. `file://hostname/path`
# NOTE: All we are interested in is the `/path` portion
RE_FILE_URI = re.compile(r"^file:(?://[^/]*)?", re.IGNORECASE)


def timestamp() -> str:
    now_utc = datetime.now(timezone.utc)
    return now_utc.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'


class SpyIO(IO[bytes]):
    stream: IO[bytes]
    spy: BinaryIO
    buf: BytesIO

    def __init__(self, stream: IO[bytes], spy: BinaryIO) -> None:
        self.stream = stream
        self.spy = spy
        self.buf = BytesIO()

    def escape(self, bs: bytes) -> bytes:
        self.buf.seek(0)
        self.buf.truncate(0)
        for b in bs:
            match b:
                case 0x00:  #-> null
                    self.buf.write(b'\\0')
                case 0x07:  #-> bell
                    self.buf.write(b'\\a')
                case 0x08:  #-> backspace
                    self.buf.write(b'\\b')
                case 0x09:  #-> horizontal tab
                    self.buf.write(b'\\t')
                case 0x0A:  #-> line feed
                    self.buf.write(b'\\n')
                case 0x0B:  #-> vertical tab
                    self.buf.write(b'\\v')
                case 0x0C:  #-> form feed
                    self.buf.write(b'\\f')
                case 0x0D:  #-> carriage return
                    self.buf.write(b'\\r')
                # case 0x1A:  #-> Control-Z
                #     self.buf.write(b'^Z')
                case 0x1B:  #-> escape
                    self.buf.write(b'\\e')
                case 0x5C:  #-> backslash
                    self.buf.write(b'\\\\')
                case _:
                    self.buf.write(b.to_bytes(2, sys.byteorder))
        return self.buf.getvalue()

    def read(self, size: int = -1, /) -> Union[bytes, Any]:
        bs = self.stream.read(size)
        self.spy.write(f'[{timestamp()}] READ: '.encode("utf-8"))
        if bs is not None and len(bs) > 0:
            self.spy.write(self.escape(bs))
        else:
            self.spy.write(b'\xe2\x90\x80')  #-> U+2400 Unicode null symbol
        self.spy.write(b'\n')
        return bs

    def write(self, b) -> int:
        self.spy.write(f'[{timestamp()}] WRITE: '.encode("utf-8"))
        num_bytes = self.stream.write(b)
        if num_bytes is not None and num_bytes > 0:
            self.spy.write(self.escape(b[:num_bytes]))
        else:
            self.spy.write(b'\xe2\x90\x80')  #-> U+2400 Unicode null symbol
        self.spy.write(b'\n')
        return num_bytes

    def flush(self) -> None:
        self.stream.flush()
        self.spy.flush()


@dataclass
class Event:
    data: Any


@dataclass
class IncomingEvent(Event):
    pass


@dataclass
class OutgoingEvent(Event):
    pass


EventPredFn = Callable[[Event], bool]
IncomingEventPredFn = Callable[[IncomingEvent], bool]
OutgoingEventPredFn = Callable[[OutgoingEvent], bool]


Callback = Callable[[Any, JsonObject], None]
ServerProcess = subprocess.Popen[bytes]


SECONDS_PER_MILLISECOND: float = 1. / 1000.


class LspTestClient(LspClient):
    server_path: Path
    server_params: List[str]
    workspace_path: Optional[Path]
    timeout_s: float

    server: ServerProcess
    process: psutil.Process
    message_stream: LspJsonStream
    ostream: IO[bytes]
    buf: BytesIO
    converter: Converter
    serial_request_id: int
    config: Dict[str, Any]
    server_capabilities: ServerCapabilities
    server_info: Optional[InitializeResultServerInfoType]
    dynamic_registrations: Dict[str, List[Registration]]
    events: List[Event]
    serial_document_id: int
    documents_by_id: Dict[int, LspTextDocument]
    documents_by_uri: Dict[str, LspTextDocument]
    active_document: Optional[LspTextDocument]
    requests_by_id: Dict[JsonValue, Any]
    responses_by_id: Dict[JsonValue, Any]
    callbacks_by_id: Dict[JsonValue, Tuple[Any, Callback]]
    stop: threading.Event
    stderr_printer: threading.Thread
    client_log_path: str
    stdout_log_path: str
    stdin_log_path: str

    def __init__(
            self,
            server_path: Path,
            server_params: List[str],
            workspace_path: Optional[Path],
            timeout_ms: float,
            config: Dict[str, Any],
            client_log_path: str,
            stdout_log_path: str,
            stdin_log_path: str
    ) -> None:
        self.server_path = server_path
        self.server_params = server_params
        self.workspace_path = workspace_path
        self.timeout_s = timeout_ms * SECONDS_PER_MILLISECOND
        self.buf = BytesIO()
        self.converter = converters.get_converter()
        self.serial_request_id = 0
        self.config = config
        self.dynamic_registrations = defaultdict(list)
        self.events = list()
        self.serial_document_id = 0
        self.documents_by_id = dict()
        self.documents_by_uri = dict()
        self.active_document = None
        self.requests_by_id = dict()
        self.responses_by_id = dict()
        self.callbacks_by_id = dict()
        self.stop = threading.Event()
        self.stderr_printer = threading.Thread(
            target=self.print_stderr,
            args=tuple()
        )
        self.client_log_path = client_log_path
        self.stdout_log_path = stdout_log_path
        self.stdin_log_path = stdin_log_path

    def print_stderr(self) -> None:
        buf = BytesIO()
        try:
            if self.server.stderr is not None:
                while self.check_server():
                    try:
                        bs = self.server.stderr.read(4096)
                        if bs is not None:
                            buf.seek(0)
                            buf.truncate(0)
                            while (bs is not None) and (len(bs) > 0):
                                buf.write(bs)
                                bs = self.server.stderr.read(4096)
                            print(buf.getvalue().decode('utf-8'), end='', file=sys.stderr)
                            continue
                    except BlockingIOError:
                        pass
                    time.sleep(0.100)
        except BaseException as e:
            print(
                "Caught unhandled exception while printing stderr logs from the server:", e,
                file=sys.stderr
            )
            traceback.print_exc(file=sys.stderr)
        finally:
            # Print whatever is left in the buffer
            buf.seek(0)
            buf.truncate(0)
            if self.server.stderr is not None:
                bs = self.server.stderr.read(4096)
                while (bs is not None) and (len(bs) > 0):
                    buf.write(bs)
                    bs = self.server.stderr.read(4096)
            print(buf.getvalue().decode('utf-8'), file=sys.stderr)

    def find_incoming_event(self, pred: IncomingEventPredFn, lower_index: int = 0) \
            -> Tuple[Optional[IncomingEvent], int]:
        upper_bound = len(self.events)
        for index in range(lower_index, upper_bound):
            event = self.events[index]
            if isinstance(event, IncomingEvent) and pred(event):
                return event, index
        return None, max(upper_bound - 1, 0)

    def has_incoming_event(self, pred: IncomingEventPredFn) -> bool:
        return self.find_incoming_event(pred)[0] is not None

    def has_outgoing_event(self, pred: OutgoingEventPredFn) -> bool:
        for event in self.events:
            if isinstance(event, OutgoingEvent) and pred(event):
                return True
        return False

    def next_request_id(self) -> int:
        request_id = self.serial_request_id
        self.serial_request_id += 1
        return request_id

    def next_document_id(self) -> int:
        document_id = self.serial_document_id
        self.serial_document_id += 1
        return document_id

    @contextmanager
    def serve(self) -> Iterator["LspTestClient"]:
        with open(self.client_log_path, "w") as log_file:
            self.log_file = log_file
            with open(self.stdout_log_path, "wb") as stdout_log:
                self.stdout_log = stdout_log
                with open(self.stdin_log_path, "wb") as stdin_log:
                    self.stdin_log = stdin_log
                    self.initialize()
                    yield self
                    self.terminate()

    def new_document(self, language_id: str) -> LspTextDocument:
        return self.open_document(language_id, None)

    def open_document(
            self,
            language_id: str,
            path: Optional[Union[Path, str]],
            make_active: bool = True
    ) -> LspTextDocument:
        document_id = self.next_document_id()
        document = LspTextDocument(self, document_id, language_id, path)
        document.on_change(self.update_document_symbols)
        self.documents_by_id[document_id] = document
        if path is not None:
            self.documents_by_uri[document.uri] = document
            document.load()
        if make_active:
            self.active_document = document
        return document

    def get_document(self, language_id: str, uri: str) -> LspTextDocument:
        document = self.documents_by_uri.get(uri, None)
        if document is not None:
            return document
        path = RE_FILE_URI.sub(uri, "")
        return self.open_document(language_id, path)

    def await_response(
            self,
            request_id: int
    ) -> Any:
        message = self.responses_by_id.get(request_id, None)
        if message is not None:
            return message
        while not self.stop.is_set():
            message = self.receive_message()
            if "result" in message:
                response_id = message.get("id", None)
                if request_id == response_id:
                    return message
            # NOTE: The response may have been recorder by a nested
            # iteration of `await_response`
            message = self.responses_by_id.get(request_id, None)
            if message is not None:
                return message

    def initialize_params(self) -> InitializeParams:
        params = InitializeParams(
            capabilities=ClientCapabilities(
                workspace=WorkspaceClientCapabilities(
                    did_change_configuration=DidChangeConfigurationClientCapabilities(
                        dynamic_registration=True,
                    ),
                    configuration=True,
                    file_operations=FileOperationClientCapabilities(
                        did_create=True,
                        will_create=True,
                        did_rename=True,
                        will_rename=True,
                        did_delete=True,
                        will_delete=True,
                    ),
                ),
                text_document=TextDocumentClientCapabilities(
                    synchronization=TextDocumentSyncClientCapabilities(
                        dynamic_registration=True,
                        will_save=True,
                        will_save_wait_until=False,
                        did_save=True,
                    ),
                ),
            ),
        )
        return params

    def initialized_params(self) -> InitializedParams:
        params = InitializedParams()
        return params

    def initialize(self) -> None:
        self.server = subprocess.Popen(
            [self.server_path] + self.server_params,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=0,
        )

        if self.server.stdout is None:
            raise RuntimeError("Cannot read from server stdout")
        if self.server.stdin is None:
            raise RuntimeError("Cannot write to server stdin")
        if self.server.stderr is None:
            raise RuntimeError("Cannot read from server stderr")

        # Make process stdout non-blocking
        stdout_fd = self.server.stdout.fileno()
        os.set_blocking(stdout_fd, False)
        stderr_fd = self.server.stderr.fileno()
        os.set_blocking(stderr_fd, False)

        self.process = psutil.Process(self.server.pid)

        self.message_stream = LspJsonStream(
            SpyIO(self.server.stdout, self.stdout_log),
            self.timeout_s,
            self.check_server
        )
        self.ostream = SpyIO(self.server.stdin, self.stdin_log)

        self.stderr_printer.start()

        initialize_id = self.send_initialize(self.initialize_params())
        self.await_response(initialize_id)

        self.send_initialized(self.initialized_params())

    def terminate(self) -> None:
        shutdown_id = self.send_shutdown()
        self.await_response(shutdown_id)

        self.send_exit()

        self.stop.set()
        self.stderr_printer.join()

        try:
            if self.timeout_s > 0.0:
                self.server.wait(timeout=self.timeout_s)
            else:
                self.server.wait()
        except subprocess.TimeoutExpired as e:
            self.kill_server()
            raise RuntimeError(
                f"Timed-out after {self.timeout_s} seconds while awaiting the server to terminate."
            ) from e

    def kill_server(self) -> None:
        self.stop.set()
        if self.server.poll() is None:
            print(
                'Server did not terminate cleanly, terminating it forcefully ...',
                file=sys.stderr
            )
            try:
                os.kill(self.server.pid, signal.SIGINT)
                self.server.wait(timeout=1.0)
            except ProcessLookupError:
                pass
            finally:
                if self.server.poll() is None:
                    try:
                        os.kill(self.server.pid, signal.SIGKILL)
                    except ProcessLookupError:
                        pass

    def check_server(self) -> bool:
        try:
            if self.process.is_running():
                if self.process.status() == psutil.STATUS_ZOMBIE:
                    self.kill_server()
                    print(
                        "ServerProcess has become defunct and will no longer respond",
                        file=sys.stderr
                    )
                else:
                    # The server is running as expected; there's nothing to do
                    pass
            else:
                self.stop.set()
                print(
                    f"ServerProcess crashed with status: {self.server.returncode}",
                    file=sys.stderr
                )
        except BaseException as e:
            self.stop.set()
            raise e
        finally:
            return not self.stop.is_set()

    def send_message(self, message: Any) -> None:
        self.events.append(OutgoingEvent(message))
        body = json.dumps(
            message,
            separators=(",",":"),
            default=self.converter.unstructure,
        ).encode("utf-8")
        self.buf.seek(0)
        self.buf.truncate(0)
        self.buf.write(f"Content-Length: {len(body)}\r\n".encode("utf-8"))
        self.buf.write(b"\r\n")
        self.buf.write(body)
        print(
            f"[{timestamp()}] Sending:\n{self.buf.getvalue().decode('utf-8')}",
            file=self.log_file
        )
        self.log_file.flush()
        if self.check_server():
            self.ostream.write(self.buf.getvalue())
            self.ostream.flush()

    def send_request(
            self,
            request_id: int,
            request: Any,
            callback: Callback
    ) -> int:
        self.requests_by_id[request_id] = request
        self.callbacks_by_id[request_id] = (request, callback)
        self.send_message(request)
        return request_id

    def receive_message(self) -> JsonObject:
        if self.check_server():
            message = self.message_stream.next()
            print(
                f"[{timestamp()}] Receiving:\n{self.message_stream.message()}",
                file=self.log_file
            )
            match message:
                case dict():
                    self.events.append(IncomingEvent(message))
                    self.dispatch_message(message)
                    return message
                case _:
                    raise RuntimeError(
                        f"Unsupported message type ({type(message)}): {message}"
                    )
        else:
            raise RuntimeError("Server has terminated.")

    def dispatch_message(self, message: JsonObject) -> None:
        if 'method' in message:
            self.dispatch_method(message)
        elif 'result' in message:
            self.dispatch_response(message)
        elif 'error' in message:
            self.dispatch_error(message)
        else:
            # notification response
            pass

    def dispatch_method(self, message: JsonObject) -> None:
        match message["method"]:
            case "client/registerCapability":
                request = self.converter.structure(
                    message,
                    ClientRegisterCapabilityRequest
                )
                response = self.receive_client_register_capability(request)
                self.send_message(response)
            case "workspace/configuration":
                request = self.converter.structure(
                    message,
                    WorkspaceConfigurationRequest
                )
                response = self.receive_workspace_configuration(request)
                self.send_message(response)
            case "telemetry/event":
                notification = self.converter.structure(
                    message,
                    TelemetryEventNotification
                )
                self.receive_telemetry_event(notification)

    def respond_to_notification(self) -> None:
        self.send_message({"id": None, "result": None, "jsonrpc": "2.0"})

    def dispatch_response(self, response: JsonObject) -> None:
        response_id = response.get("id", None)
        if response_id is not None:
            self.responses_by_id[response_id] = response
            record = self.callbacks_by_id.get(response_id, None)
            if record is not None:
                request, callback = record
                callback(request, response)
            else:
                raise ValueError(
                    f'No requests exists for response Id: {response_id}'
                )

    def dispatch_error(self, message: JsonObject) -> None:
        raise RuntimeError(
            f'ServerProcess responded with an error: {message}'
        )

    def serialize_json(self, message: Any) -> str:
        return json.dumps(message, default=self.converter.unstructure)

    def receive_telemetry_event(self, notification: TelemetryEventNotification) -> None:
        # Check the event in the logs
        pass

    def send_initialize(self, params: InitializeParams) -> int:
        request_id = self.next_request_id()
        request = InitializeRequest(request_id, params)
        return self.send_request(request_id, request, self.receive_initialize)

    def receive_initialize(self, request: Any, message: JsonObject) -> None:
        response = self.converter.structure(
            message,
            InitializeResponse
        )
        result = response.result
        self.server_capabilities = result.capabilities
        self.server_info = result.server_info

    def send_initialized(self, params: InitializedParams) -> None:
        notification = InitializedNotification(params)
        self.send_message(notification)

    def send_shutdown(self) -> int:
        request_id = self.next_request_id()
        request = ShutdownRequest(request_id)
        return self.send_request(request_id, request, self.receive_shutdown)

    def receive_shutdown(self, request: Any, message: JsonObject) -> None:
        # response = self.converter.structure(
        #     message,
        #     ShutdownResponse
        # )
        pass

    def send_exit(self) -> None:
        notification = ExitNotification()
        self.send_message(notification)

    def receive_client_register_capability(
            self,
            request: ClientRegisterCapabilityRequest
    ) -> ClientRegisterCapabilityResponse:
        for registration in request.params.registrations:
            registrations = self.dynamic_registrations[registration.method]
            registrations.append(registration)
            match registration.method:
                case "workspace/didChangeConfiguration":
                    self.send_workspace_did_change_configuration(
                        DidChangeConfigurationParams(self.config)
                    )
        response = ClientRegisterCapabilityResponse(request.id)
        return response

    def send_workspace_did_change_configuration(
            self,
            params: DidChangeConfigurationParams
    ) -> None:
        notification = WorkspaceDidChangeConfigurationNotification(params)
        self.send_message(notification)

    def receive_workspace_configuration(
            self,
            request: WorkspaceConfigurationRequest
    ) -> WorkspaceConfigurationResponse:
        configs = []
        for item in request.params.items:
            config: Any = self.config
            if item.section is not None:
                for name in item.section.split("."):
                    config = config.get(name, None)
                    if config is None:
                        break
            configs.append(config)
        response = WorkspaceConfigurationResponse(request.id, configs)
        return response

    @requires_server_capabilities
    def server_supports_text_document_did_open_or_close(self) -> bool:
        text_document_sync = self.server_capabilities.text_document_sync
        if isinstance(text_document_sync, TextDocumentSyncOptions):
            return bool(text_document_sync.open_close)
        return False

    def send_text_document_did_open(
            self,
            params: DidOpenTextDocumentParams
    ) -> None:
        notification = TextDocumentDidOpenNotification(params)
        self.send_message(notification)

    def text_document_did_open(
            self,
            document_id: int,
            uri: str,
            language_id: str,
            version: int,
            text: str
    ) -> None:
        document = self.documents_by_id[document_id]
        self.documents_by_uri[uri] = document
        if self.server_supports_text_document_did_open_or_close():
            params = DidOpenTextDocumentParams(
                text_document=TextDocumentItem(
                    uri=uri,
                    language_id=language_id,
                    version=version,
                    text=text
                )
            )
            self.send_text_document_did_open(params)

    def send_text_document_did_close(
            self,
            params: DidCloseTextDocumentParams
    ) -> None:
        notification = TextDocumentDidCloseNotification(params)
        self.send_message(notification)

    def text_document_did_close(
            self,
            document_id: int,
            uri: Optional[str]
    ) -> None:
        del self.documents_by_id[document_id]
        if self.active_document is not None \
           and self.active_document.document_id == document_id:
            for prev_document_id in range(document_id - 1, -1, -1):
                prev_document = self.documents_by_id.get(prev_document_id, None)
                if prev_document is not None:
                    self.active_document = prev_document
        if uri is not None:
            del self.documents_by_uri[uri]
            if self.server_supports_text_document_did_open_or_close():
                params = DidCloseTextDocumentParams(
                    text_document=TextDocumentIdentifier(
                        uri=uri
                    )
                )
                self.send_text_document_did_close(params)

    def send_text_document_will_save(
            self,
            params: WillSaveTextDocumentParams
    ) -> None:
        notification = TextDocumentWillSaveNotification(params)
        self.send_message(notification)

    @requires_server_capabilities
    def server_supports_text_document_will_save(self) -> bool:
        text_document_sync = self.server_capabilities.text_document_sync
        if isinstance(text_document_sync, TextDocumentSyncOptions):
            return bool(text_document_sync.will_save)
        return False

    def text_document_will_save(
            self,
            uri: str,
            reason: TextDocumentSaveReason
    ) -> None:
        if self.server_supports_text_document_will_save():
            params = WillSaveTextDocumentParams(
                text_document=TextDocumentIdentifier(
                    uri=uri
                ),
                reason=reason
            )
            self.send_text_document_will_save(params)

    def send_text_document_will_save_wait_until(
            self,
            params: WillSaveTextDocumentParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentWillSaveWaitUntilRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_will_save_wait_until
        )

    def receive_text_document_will_save_wait_until(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        response = self.converter.structure(
            message,
            TextDocumentWillSaveWaitUntilResponse
        )
        if response.result is not None:
            uri = request.uri
            document = self.documents_by_uri[uri]
            document.apply(response.result)

    @requires_server_capabilities
    def server_supports_text_document_will_save_wait_until(self) -> bool:
        text_document_sync = self.server_capabilities.text_document_sync
        if isinstance(text_document_sync, TextDocumentSyncOptions):
            return bool(text_document_sync.will_save_wait_until)
        return False

    def text_document_will_save_wait_until(
            self,
            uri: str,
            reason: TextDocumentSaveReason
    ) -> None:
        if self.server_supports_text_document_will_save_wait_until():
            params = WillSaveTextDocumentParams(
                text_document=TextDocumentIdentifier(
                    uri=uri
                ),
                reason=reason
            )
            request_id = self.send_text_document_will_save_wait_until(params)
            self.await_response(request_id)

    def send_text_document_did_save(
            self,
            params: DidSaveTextDocumentParams
    ) -> None:
        notification = TextDocumentDidSaveNotification(params)
        self.send_message(notification)

    @requires_server_capabilities
    def server_supports_full_text_on_save(self) -> bool:
        text_document_sync = self.server_capabilities.text_document_sync
        if isinstance(text_document_sync, TextDocumentSyncOptions):
            save = text_document_sync.save
            if isinstance(save, SaveOptions):
                return bool(save.include_text)
        return False

    @requires_server_capabilities
    def server_supports_did_save(self) -> bool:
        text_document_sync = self.server_capabilities.text_document_sync
        if isinstance(text_document_sync, TextDocumentSyncOptions):
            return text_document_sync.save is not None
        return False

    def text_document_did_save(self, uri: str, text: str) -> None:
        if self.server_supports_did_save():
            params = DidSaveTextDocumentParams(
                text_document=TextDocumentIdentifier(
                    uri=uri
                ),
                text=text if self.server_supports_full_text_on_save() else None
            )
            self.send_text_document_did_save(params)

    def send_text_document_did_change(
            self,
            params: DidChangeTextDocumentParams
    ) -> None:
        notification = TextDocumentDidChangeNotification(params)
        self.send_message(notification)

    @requires_server_capabilities
    def server_supports_text_document_sync_kind(
            self,
            kind: TextDocumentSyncKind
    ) -> bool:
        text_document_sync = self.server_capabilities.text_document_sync
        if text_document_sync is not None:
            match text_document_sync:
                case TextDocumentSyncKind():
                    return kind == text_document_sync
                case TextDocumentSyncOptions():
                    if text_document_sync.change is not None:
                        return kind == text_document_sync.change
        return kind == TextDocumentSyncKind.Full

    @requires_server_capabilities
    def server_supports_text_document_did_change(self) -> bool:
        text_document_sync = self.server_capabilities.text_document_sync
        if isinstance(text_document_sync, TextDocumentSyncOptions):
            return text_document_sync.change is not None \
                and text_document_sync.change != TextDocumentSyncKind.None_
        return False

    def text_document_did_change(
            self,
            uri: str,
            version: int,
            start_line: int,
            start_column: int,
            end_line: int,
            end_column: int,
            full_text: str,
            part_text: str
    ) -> None:
        if self.server_supports_text_document_did_change():
            params = DidChangeTextDocumentParams(
                text_document=VersionedTextDocumentIdentifier(
                    version=version,
                    uri=uri
                ),
                content_changes=[
                    TextDocumentContentChangeEvent_Type1(
                        range=Range(
                            start=Position(
                                line=start_line,
                                character=start_column
                            ),
                            end=Position(
                                line=end_line,
                                character=end_column
                            )
                        ),
                        text=part_text
                    ) \
                    if self.server_supports_text_document_sync_kind(
                        TextDocumentSyncKind.Incremental
                    ) \
                    else TextDocumentContentChangeEvent_Type2(
                        text=full_text
                    )
                ]
            )
            self.send_text_document_did_change(params)

    @requires_server_capabilities
    def server_supports_workspace_will_create_files(self) -> bool:
        workspace = self.server_capabilities.workspace
        if workspace is not None:
            file_operations = workspace.file_operations
            if file_operations is not None:
                # TODO: Implement a parser for additional pattern matching
                return file_operations.will_create is not None
        return False

    def send_workspace_will_create_files(self, params: CreateFilesParams) -> int:
        request_id = self.next_request_id()
        request = WorkspaceWillCreateFilesRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_workspace_will_create_files
        )

    def receive_workspace_will_create_files(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # TODO: Update the workspace as requested by the response
        pass

    def workspace_will_create_files(self, files: List[Uri]) -> None:
        if self.server_supports_workspace_will_create_files():
            params = CreateFilesParams(
                files=[FileCreate(uri) for uri in files]
            )
            request_id = self.send_workspace_will_create_files(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_workspace_did_create_files(self) -> bool:
        workspace = self.server_capabilities.workspace
        if workspace is not None:
            file_operations = workspace.file_operations
            if file_operations is not None:
                # TODO: Implement a parser for additional pattern matching
                return file_operations.did_create is not None
        return False

    def send_workspace_did_create_files(self, params: CreateFilesParams) -> None:
        notification = WorkspaceDidCreateFilesNotification(params)
        self.send_message(notification)

    def workspace_did_create_files(self, files: List[Uri]) -> None:
        for uri in files:
            if uri in self.documents_by_uri:
                del self.documents_by_uri[uri]
        if self.server_supports_workspace_did_create_files():
            params = CreateFilesParams(
                files=[FileCreate(uri) for uri in files]
            )
            self.send_workspace_did_create_files(params)

    @requires_server_capabilities
    def server_supports_workspace_will_rename_files(self) -> bool:
        workspace = self.server_capabilities.workspace
        if workspace is not None:
            file_operations = workspace.file_operations
            if file_operations is not None:
                # TODO: Implement a parser for additional pattern matching
                return file_operations.will_rename is not None
        return False

    def send_workspace_will_rename_files(self, params: RenameFilesParams) -> int:
        request_id = self.next_request_id()
        request = WorkspaceWillRenameFilesRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_workspace_will_rename_files
        )

    def receive_workspace_will_rename_files(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # TODO: Update the workspace as requested by the response
        pass

    def workspace_will_rename_files(self, files: List[FileRenameMapping]) -> None:
        if self.server_supports_workspace_will_rename_files():
            params = RenameFilesParams(
                files=[FileRename(old_uri, new_uri) for old_uri, new_uri in files]
            )
            request_id = self.send_workspace_will_rename_files(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_workspace_did_rename_files(self) -> bool:
        workspace = self.server_capabilities.workspace
        if workspace is not None:
            file_operations = workspace.file_operations
            if file_operations is not None:
                # TODO: Implement a parser for additional pattern matching
                return file_operations.did_rename is not None
        return False

    def send_workspace_did_rename_files(self, params: RenameFilesParams) -> None:
        notification = WorkspaceDidRenameFilesNotification(params)
        self.send_message(notification)

    def workspace_did_rename_files(self, files: List[FileRenameMapping]) -> None:
        for old_uri, new_uri in files:
            document = self.documents_by_uri[old_uri]
            del self.documents_by_uri[old_uri]
            self.documents_by_uri[new_uri] = document
        if self.server_supports_workspace_did_rename_files():
            params = RenameFilesParams(
                files=[FileRename(old_uri, new_uri) for old_uri, new_uri in files]
            )
            self.send_workspace_did_rename_files(params)

    @requires_server_capabilities
    def server_supports_workspace_will_delete_files(self) -> bool:
        workspace = self.server_capabilities.workspace
        if workspace is not None:
            file_operations = workspace.file_operations
            if file_operations is not None:
                # TODO: Implement a parser for additional pattern matching
                return file_operations.will_delete is not None
        return False

    def send_workspace_will_delete_files(self, params: DeleteFilesParams) -> int:
        request_id = self.next_request_id()
        request = WorkspaceWillDeleteFilesRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_workspace_will_delete_files
        )

    def receive_workspace_will_delete_files(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # TODO: Update the workspace as requested by the response
        pass

    def workspace_will_delete_files(self, files: List[Uri]) -> None:
        if self.server_supports_workspace_will_delete_files():
            params = DeleteFilesParams(
                files=[FileDelete(uri) for uri in files]
            )
            request_id = self.send_workspace_will_delete_files(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_workspace_did_delete_files(self) -> bool:
        workspace = self.server_capabilities.workspace
        if workspace is not None:
            file_operations = workspace.file_operations
            if file_operations is not None:
                # TODO: Implement a parser for additional pattern matching
                return file_operations.did_delete is not None
        return False

    def send_workspace_did_delete_files(self, params: DeleteFilesParams) -> None:
        notification = WorkspaceDidDeleteFilesNotification(params)
        self.send_message(notification)

    def workspace_did_delete_files(self, files: List[Uri]) -> None:
        for uri in files:
            if uri in self.documents_by_uri:
                del self.documents_by_uri[uri]
        if self.server_supports_workspace_did_delete_files():
            params = DeleteFilesParams(
                files=[FileDelete(uri) for uri in files]
            )
            self.send_workspace_did_delete_files(params)

    @requires_server_capabilities
    def server_supports_document_highlight(self) -> bool:
        return self.server_capabilities.document_highlight_provider is not None

    def send_text_document_document_highlight(
            self,
            params: DocumentHighlightParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentDocumentHighlightRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_document_highlight
        )

    def receive_text_document_document_highlight(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # NOTE: Implement this in the concrete subclass because it may require
        # language-specific information (such as the language identifier for the
        # associated text document).
        raise NotImplementedError

    def highlight_symbol(self, uri: str, line: int, column: int) -> None:
        if self.server_supports_document_highlight():
            params = DocumentHighlightParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
                position=Position(
                    line=line,
                    character=column,
                ),
            )
            request_id = self.send_text_document_document_highlight(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_document_hover(self) -> bool:
        return self.server_capabilities.hover_provider is not None

    def send_text_document_hover(
            self,
            params: HoverParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentHoverRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_hover
        )

    def receive_text_document_hover(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # NOTE: Implement this in the concrete subclass because it may require
        # language-specific information (such as the language identifier for the
        # associated text document).
        raise NotImplementedError

    def hover(self, uri: str, line: int, column: int) -> None:
        if self.server_supports_document_hover():
            params = HoverParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
                position=Position(
                    line=line,
                    character=column,
                ),
            )
            request_id = self.send_text_document_hover(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_document_symbols(self) -> bool:
        return self.server_capabilities.document_symbol_provider is not None

    def send_text_document_document_symbol(
            self,
            params: DocumentSymbolParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentDocumentSymbolRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_document_symbol
        )

    def receive_text_document_document_symbol(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # NOTE: Implement this in the concrete subclass because it may require
        # language-specific information (such as the language identifier for the
        # associated text document).
        raise NotImplementedError

    def update_document_symbols(self, document_id: int, uri: Optional[str]) -> None:
        if (uri is not None) and self.server_supports_document_symbols():
            params = DocumentSymbolParams(
                text_document=TextDocumentIdentifier(
                    uri=uri
                )
            )
            request_id = self.send_text_document_document_symbol(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_semantic_highlight(self) -> bool:
        return self.server_capabilities.semantic_tokens_provider is not None

    def send_text_document_semantic_tokens_full(
            self,
            params: SemanticTokensParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentSemanticTokensFullRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_semantic_tokens_full
        )

    def receive_text_document_semantic_tokens_full(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # NOTE: Implement this in the concrete subclass because it may require
        # language-specific information (such as the language identifier for the
        # associated text document).
        raise NotImplementedError

    def semantic_highlight(self, uri: str) -> None:
        if self.server_supports_document_highlight():
            params = SemanticTokensParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
            )
            request_id = self.send_text_document_semantic_tokens_full(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_code_completion(self) -> bool:
        return self.server_capabilities.completion_provider is not None

    def send_text_document_completion(
            self,
            params: CompletionParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentCompletionRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_completion
        )

    def receive_text_document_completion(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # NOTE: Implement this in the concrete subclass because it may require
        # language-specific information (such as the language identifier for the
        # associated text document).
        raise NotImplementedError

    def complete(self, uri: str, line: int, column: int) -> None:
        if self.server_supports_document_highlight():
            params = CompletionParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
                position=Position(
                    line=line,
                    character=column,
                ),
            )
            request_id = self.send_text_document_completion(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_formatting(self) -> bool:
        return self.server_capabilities.document_formatting_provider is not None

    def send_text_document_formatting(
            self,
            params: DocumentFormattingParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentFormattingRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_formatting
        )

    def receive_text_document_formatting(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # NOTE: Implement this in the concrete subclass because it may require
        # language-specific information (such as the language identifier for the
        # associated text document).
        raise NotImplementedError

    def format(self, uri: str) -> None:
        if self.server_supports_formatting():
            params = DocumentFormattingParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
                options=FormattingOptions(
                    tab_size=4,
                    insert_spaces=True,
                ),
            )
            request_id = self.send_text_document_formatting(params)
            self.await_response(request_id)

    @requires_server_capabilities
    def server_supports_range_formatting(self) -> bool:
        return self.server_capabilities.document_range_formatting_provider is not None

    def send_text_document_range_formatting(
            self,
            params: DocumentRangeFormattingParams
    ) -> int:
        request_id = self.next_request_id()
        request = TextDocumentRangeFormattingRequest(request_id, params)
        return self.send_request(
            request_id,
            request,
            self.receive_text_document_range_formatting
        )

    def receive_text_document_range_formatting(
            self,
            request: Any,
            message: JsonObject
    ) -> None:
        # NOTE: Implement this in the concrete subclass because it may require
        # language-specific information (such as the language identifier for the
        # associated text document).
        raise NotImplementedError

    def format_range(self, uri: str, selection: Range) -> None:
        if self.server_supports_range_formatting():
            params = DocumentRangeFormattingParams(
                text_document=TextDocumentIdentifier(
                    uri=uri,
                ),
                range=selection,
                options=FormattingOptions(
                    tab_size=4,
                    insert_spaces=True,
                ),
            )
            request_id = self.send_text_document_range_formatting(params)
            self.await_response(request_id)



import sys
import traceback
from collections import defaultdict
from contextlib import contextmanager
from functools import wraps
from pathlib import Path
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Union

from llanguage_server.file_generator import FileGenerator
from llanguage_server.lsp.datatypes import LspSpec, LspSymbol, LspSymbolKind
from llanguage_server.lsp.utils import (as_enumeration_spec,
                                        method_to_camel_case, rename_enum,
                                        rename_field, rename_type, upper_first)
from llanguage_server.lsp.visitors import LspAnalysisPipeline, NestedContext


def gensym_context(fn: Callable) -> Callable:
    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        old_serial_id_by_sym = self.serial_id_by_sym
        self.serial_id_by_sym = defaultdict(int)
        try:
            return fn(self, *args, **kwargs)
        finally:
            self.serial_id_by_sym = old_serial_id_by_sym
    return wrapper


class CPlusPlusLspFileGenerator(FileGenerator, NestedContext):
    schema: LspSpec
    pipeline: LspAnalysisPipeline
    namespace: str

    serial_id_by_sym: Dict[str, int] = defaultdict(int)

    _type_declaration_wrap_ptr: bool = False
    _type_declaration_wrap_nested_ptr: bool = False
    _type_declaration_string_ptr: bool = False
    _type_declaration_nested_string_ptr: bool = False

    def __init__(
            self,
            output_path: Path,
            schema: LspSpec,
            pipeline: LspAnalysisPipeline,
            namespace: str
    ) -> None:
        super().__init__(output_path)
        self.schema = schema
        self.pipeline = pipeline
        self.namespace = namespace

    def gensym(self, prefix: str = "tmp") -> str:
        next_id = self.serial_id_by_sym[prefix]
        self.serial_id_by_sym[prefix] += 1
        return f'{prefix}_{next_id}'

    def gensym_decl(self, kind_or_spec: Union[str, LspSpec], nym: str) -> str:
        kind: str
        match kind_or_spec:
            case str():
                kind = kind_or_spec
            case dict():
                spec: LspSpec = kind_or_spec
                kind = self.get_type_declaration(spec)
        sym = self.gensym(nym)
        self.write(f'{kind} {sym};')
        return sym

    def gensym_ref(self, nym: str, expr: str, const: bool = True) -> str:
        sym = self.gensym(nym)
        if const:
            self.gen_assign(f'const auto &{sym}', expr)
        else:
            self.gen_assign(f'auto &{sym}', expr)
        return sym

    def gensym_init(self, nym: str, expr: str) -> str:
        sym = self.gensym(nym)
        self.gen_assign(f'auto {sym}', expr)
        return sym

    def gensym_array(
            self,
            array_type_or_spec: Union[str, LspSpec],
            prefix: str,
            size_expr: Optional[str] = None
    ) -> str:
        array_type: str
        match array_type_or_spec:
            case str():
                array_type = array_type_or_spec
            case dict():
                array_spec: LspSpec = array_type_or_spec
                array_type = self.get_type_declaration(array_spec)
        name = self.gensym(prefix)
        self.write(f'{array_type} {name};')
        if size_expr is not None:
            self.write(f'{name}.reserve({size_expr});')
        return name

    def type_name_or_spec_to_decl(
            self,
            type_name_or_spec: Union[str, LspSpec]
    ) -> str:
        match type_name_or_spec:
            case str():
                return type_name_or_spec
            case dict():
                type_spec: LspSpec = type_name_or_spec
                return self.get_type_declaration(type_spec)
            case _:
                raise ValueError(
                    'Unsupported type ({kind}): {spec}'.format(
                        kind=type(type_name_or_spec),
                        spec=type_name_or_spec,
                    )
                )

    @contextmanager
    def gensym_foreach(
            self,
            coll_name: str,
            elem_prefix: str = 'elem'
    ) -> Iterator[str]:
        elem_name = self.gensym(elem_prefix)
        with self.gen_foreach(f'const auto &{elem_name}', coll_name):
            yield elem_name

    @contextmanager
    def gensym_foreach_keyval(
            self,
            map_name: str,
            key_prefix: str = 'key',
            value_prefix: str = 'value',
            const: bool = True
    ) -> Iterator[Tuple[str, str]]:
        key_name = self.gensym(key_prefix)
        value_name = self.gensym(value_prefix)
        prefix = 'const ' if const else ''
        with self.gen_foreach(f'{prefix}auto &[{key_name}, {value_name}]', map_name):
            yield key_name, value_name

    @contextmanager
    def block(
            self,
            indent: bool = True,
            end: str = '\n'
    ) -> Iterator:
        self.inline('{', indent=indent, end='\n')
        with self.indent():
            yield self
        self.inline('}', indent=True, end=end)

    def gen_include(self, header: str) -> None:
        self.write(f'#include <{header}>')

    def pragma_once(self) -> None:
        self.write('#pragma once')

    @contextmanager
    def gen_namespace(
            self,
            namespace: str
    ) -> Iterator:
        self.write(f'namespace {namespace} {{')
        with self.indent():
            yield self
        self.write(f'}} // namespace {namespace}')

    @contextmanager
    def gen_class(
            self,
            class_name: str,
            sups: Optional[List[Union[LspSpec, str]]] = None
    ) -> Iterator:
        if sups is None:
            self.write(f'class {class_name} {{')
        else:
            sups = [
                sup_spec["name"] if isinstance(sup_spec, dict) else sup_spec
                for sup_spec in sups
            ]
            self.write(f'class {class_name}')
            with self.indent():
                sup_name = sups[0]
                self.write(f': public {sup_name}')
                for i in range(1, len(sups)):
                    sup_name = sups[i]
                    self.write(f', public {sup_name}')
            self.write('{')
        yield self
        self.write(f'}}; // class {class_name}')

    @contextmanager
    def gen_public(self) -> Iterator:
        self.write('public:')
        with self.indent():
            yield self

    @contextmanager
    def gen_protected(self) -> Iterator:
        self.write('protected:')
        with self.indent():
            yield self

    @contextmanager
    def gen_private(self) -> Iterator:
        self.write('private:')
        with self.indent():
            yield self

    @contextmanager
    def gen_struct(
            self,
            struct_name: str,
            sups: Optional[List[LspSpec]] = None
    ) -> Iterator:
        if sups is None:
            self.write(f'struct {struct_name} {{')
        else:
            self.write(f'struct {struct_name}')
            with self.indent():
                sup_spec = sups[0]
                self.write(f': public {sup_spec["name"]}')
                for i in range(1, len(sups)):
                    sup_spec = sups[i]
                    self.write(f', public {sup_spec["name"]}')
            self.write('{')
        with self.indent():
            yield self
        self.write(f'}}; // struct {struct_name}')

    @contextmanager
    def gen_union(self) -> Iterator:
        self.write(f'union {{')
        with self.indent():
            yield self
        self.write('};')

    def gen_union_with(self, item_specs: List[LspSpec]) -> None:
        with self.gen_union():
            for item_index, item_spec in enumerate(item_specs):
                with self.nest_name(str(item_index)):
                    self.gen_union_field(item_spec)

    def gen_union_field(self, item_spec: LspSpec) -> None:
        with self.type_declaration_options(string_ptr=True):
            field_name = self.name_field(item_spec)
            field_type = self.get_type_declaration(item_spec)
            self.write(f'{field_type} _{field_name};')

    @contextmanager
    def gen_enum(self, name: str) -> Iterator:
        self.write(f'enum class {name} {{')
        with self.indent():
            yield self
        self.write(f'}}; // enum class {name}')

    def gen_enum_with(self, enum_name: str, item_specs: List[LspSpec]) -> None:
        with self.gen_enum(enum_name):
            self.write('Uninitialized = -1,')
            for index, item_spec in enumerate(item_specs):
                with self.nest_name(str(index)):
                    self.write(f'{self.get_union_enumeration(item_spec)} = {index},')

    def gen_union_tag_field(self, enum_name: str) -> None:
        line = f'{self.indentation()}{enum_name} _type{{{enum_name}::Uninitialized}};'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            self.write(f'{enum_name} _type{{')
            with self.indent():
                self.write(f'{enum_name}::Uninitialized')
            self.write('};')

    @contextmanager
    def gen_switch(self, value: str) -> Iterator:
        self.write(f'switch ({value}) {{')
        yield self
        self.write('}')

    def gen_map_decl(self, map_name: str, key_type: str, val_type: str) -> None:
        line = f'{self.indentation()}extern const std::map<{key_type}, {val_type}> {map_name};'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            line = f'{self.indentation()}extern const std::map<{key_type}, {val_type}>'
            if len(line) < self.column_width:
                self.inline(line, end='\n')
                with self.indent():
                    self.write(f'{map_name};')
            else:
                self.write('extern const std::map<')
                with self.indent():
                    self.write(f'{key_type},')
                    self.write(val_type)
                self.write(f'> {map_name};')

    @contextmanager
    def gen_map(
            self,
            map_name: str,
            key_type: str,
            val_type: str
    ) -> Iterator:
        line = f'{self.indentation()}const std::map<{key_type}, {val_type}> {map_name} = {{'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            line = f'{self.indentation()}const std::map<{key_type}, {val_type}>'
            if len(line) < self.column_width:
                self.inline(line, end='\n')
                self.write(f'{map_name} = {{')
            else:
                self.write('const std::map<')
                with self.indent():
                    self.write(f'{key_type},')
                    self.write(val_type)
                self.write(f'> {map_name} = {{')
        with self.indent():
            yield self
        self.write('};')

    def gen_map_entry(self, key: str, value: str) -> None:
        line = f'{self.indentation()}{{{key}, {value}}},'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            self.write(f'{{{key},')
            self.write(f' {value}}},')

    def gen_enum_map_entry(
            self,
            enumeration: str,
            enumerator_or_spec: Union[str, LspSpec],
            value_or_spec: Union[str, LspSpec]
    ) -> None:
        match (enumerator_or_spec, value_or_spec):
            case (str(), str()):
                enumerator: str = enumerator_or_spec
                key = f'{enumeration}::{enumerator}'
                value = f'"{value_or_spec}"'
            case (str(), dict()):
                enumerator: str = enumerator_or_spec
                value_spec: LspSpec = value_or_spec
                key = f'{enumeration}::{enumerator}'
                value = '"{enum_value}"'.format(
                    enum_value=self.get_union_enumeration(value_spec)
                )
            case (dict(), str()):
                enumerator_spec: LspSpec = enumerator_or_spec
                key = f'{enumeration}::{{enumerator}}'.format(
                    enumerator=self.get_union_enumeration(enumerator_spec)
                )
                value = f'"{value_or_spec}"'
            case (dict(), dict()):
                enumerator_spec = enumerator_or_spec
                value_spec = value_or_spec
                key = f'{enumeration}::{{enumerator}}'.format(
                    enumerator=self.get_union_enumeration(enumerator_spec)
                )
                value = '"{enum_value}"'.format(
                    enum_value=self.get_union_enumeration(value_spec)
                )
            case _:
                raise ValueError(
                    f'Unsupported enumerator type ({type(enumerator_or_spec)}) or ' +
                    f'value type ({type(value_or_spec)})'
                )
        self.gen_map_entry(key, value)

    @contextmanager
    def gen_case(
            self,
            enum_name: str,
            spec_or_name: Union[LspSpec, str]
    ) -> Iterator:
        enumerator = self.get_union_enumeration(spec_or_name)
        self.inline(f'case {enum_name}::{rename_enum(enumerator)}: ', indent=True)
        with self.block(indent=False):
            yield self

    @contextmanager
    def gen_default(self) -> Iterator:
        self.inline('default: ', indent=True)
        with self.block(indent=False):
            yield self

    @contextmanager
    def gen_try(self) -> Iterator:
        self.write('try {')
        with self.indent():
            yield self
        self.inline('}', indent=True, end='')

    @contextmanager
    def gen_catch(self, expr: str, end: str = '\n') -> Iterator:
        self.inline(f' catch ({expr}) {{', end='\n')
        with self.indent():
            yield self
        self.inline('}', indent=True, end=end)

    @contextmanager
    def gen_if(self, expr: str, end: str = '\n') -> Iterator:
        self.write(f'if ({expr}) {{')
        with self.indent():
            yield self
        self.inline('}', indent=True, end=end)

    @contextmanager
    def gen_if_eq(self, lhs: str, rhs: str, end: str = '\n') -> Iterator:
        line = f'{self.indentation()}if ({lhs} == {rhs}) {{'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            self.write(f'if ({lhs} ==')
            with self.indent(2):
                self.write(f'{rhs}) {{')
        with self.indent():
            yield self
        self.inline('}', indent=True, end=end)

    @contextmanager
    def gen_if_ne(self, lhs: str, rhs: str, end: str = '\n') -> Iterator:
        line = f'{self.indentation()}if ({lhs} != {rhs}) {{'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            self.write(f'if ({lhs} !=')
            with self.indent(2):
                self.write(f'{rhs}) {{')
        with self.indent():
            yield self
        self.inline('}', indent=True, end=end)

    @contextmanager
    def gen_elif(self, expr: str, end: str = '\n') -> Iterator:
        self.inline(f' else if ({expr}) {{', end='\n')
        with self.indent():
            yield self
        self.inline('}', indent=True, end=end)

    @contextmanager
    def gen_else(self) -> Iterator:
        self.inline(' else {', end='\n')
        with self.indent():
            yield self
        self.write('}')

    @contextmanager
    def gen_for(self, init: str, pred: str, step: str) -> Iterator:
        line = f'{self.indentation()}for ({init}; {pred}; {step}) {{'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            self.write(f'for (')
            with self.indent():
                self.write(f'{init};')
                self.write(f'{pred};')
                self.write(f'{step}')
            self.write(') {')
        with self.indent():
            yield self
        self.write('}')

    @contextmanager
    def gen_foreach(self, elem: str, coll: str) -> Iterator:
        line = f'{self.indentation()}for ({elem} : {coll}) {{'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            self.write(f'for ({elem}')
            with self.indent(2):
                self.write(f': {coll}) {{')
        with self.indent():
            yield self
        self.write('}')

    @contextmanager
    def gen_while(self, pred: str) -> Iterator:
        self.write(f'while ({pred}) {{')
        with self.indent():
            yield self
        self.write('}')

    @contextmanager
    def gen_do(self, pred: str) -> Iterator:
        self.write('do {')
        with self.indent():
            yield self
        self.write(f'}} while ({pred});')

    def gen_assign(self, lhs: str, rhs: str) -> None:
        line = f'{self.indentation()}{lhs} = {rhs};'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            self.write(f'{lhs} =')
            with self.indent():
                self.write(f'{rhs};')

    def gen_assign_enum(
            self,
            lhs: str,
            enumeration: str,
            enumerator: Union[str, LspSpec]
    ) -> None:
        enumerator = self.get_union_enumeration(enumerator)
        self.gen_assign(lhs, f'{enumeration}::{enumerator}')

    def gen_fn_decl(
            self,
            fn: str,
            ret_type: str = 'void',
            params: List[Union[str, LspSpec]] = [],
            specs: Optional[Union[str, List[str]]] = None,
            docs: Optional[str] = None,
            virtual: bool = False,
            pure: bool = False,
            override: bool = False
    ) -> None:
        self.generate_docstring(docs)
        params = [(self.get_type_declaration(param)
                   if isinstance(param, dict)
                   else param)
                  for param in params]
        prefix = 'virtual ' if virtual else ''
        suffix = ' override' if override else ''
        line = f'{self.indentation()}{prefix}auto {fn}({", ".join(params)})'
        if specs is not None:
            if isinstance(specs, list):
                specs = " ".join(specs)
            line += f' {specs}'
        line += f' -> {ret_type}{suffix}'
        if pure:
            line += ' = 0'
        line += ';'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            if len(params) == 0:
                self.inline(f'{prefix}auto {fn}()', indent=True)
                if specs is not None:
                    self.inline(f' {specs}')
                self.inline(end='\n')
                self.inline(indent=True)
            else:
                self.write(f'{prefix}auto {fn}(')
                with self.indent():
                    self.inline(params[0], indent=True)
                    for i in range(1, len(params)):
                        self.inline(',', end='\n')
                        self.inline(params[i], indent=True)
                    self.inline(end='\n')
                self.inline(') ', indent=True)
                if specs is not None:
                    self.inline(f'{specs} ')
            self.inline(f'-> {ret_type}{suffix}')
            if pure:
                self.inline(' = 0')
            self.inline(';', end='\n')

    @contextmanager
    def gen_fn(
            self,
            fn: str,
            ret_type_or_spec: Union[str, LspSpec] = 'void',
            params: List[Union[str, LspSpec]] = [],
            specs: Optional[str] = None,
            inline: bool = False
    ) -> Iterator:
        ret_type: str
        match ret_type_or_spec:
            case str():
                ret_type = ret_type_or_spec
            case dict():
                type_spec: LspSpec = ret_type_or_spec
                ret_type = self.get_type_declaration(type_spec)
        params = [(self.get_type_declaration(param)
                   if isinstance(param, dict)
                   else param)
                  for param in params]
        if inline:
            prefix = 'inline '
        else:
            prefix = ''
        line = f'{self.indentation()}{prefix}auto {fn}({", ".join(params)})'
        if specs is not None:
            if isinstance(specs, list):
                specs = " ".join(specs)
            line += f' {specs}'
        line += f' -> {ret_type}'
        if len(line) < self.column_width:
            self.inline(line, end='\n')
        else:
            if len(params) == 0:
                self.inline(f'{prefix}auto {fn}()', indent=True)
                if specs is not None:
                    self.inline(f' {specs}')
                self.inline(end='\n')
                self.inline(indent=True)
            else:
                self.write(f'{prefix}auto {fn}(')
                with self.indent():
                    self.inline(params[0], indent=True)
                    for i in range(1, len(params)):
                        self.inline(',', end='\n')
                        self.inline(params[i], indent=True)
                    self.inline(end='\n')
                self.inline(') ', indent=True)
                if specs is not None:
                    self.inline(f'{specs} ')
            self.inline(f'-> {ret_type}', end='\n')
        with self.block():
            yield self

    def gen_constructor_decl(
            self,
            kind: str,
            params: List[Union[str, LspSpec]] = [],
            specs: Optional[str] = None
    ) -> None:
        params = [(self.get_type_declaration(param)
                   if isinstance(param, dict)
                   else param)
                  for param in params]
        line = f'{self.indentation()}{kind}({", ".join(params)})'
        if specs is not None:
            if isinstance(specs, list):
                specs = " ".join(specs)
            line += f' {specs}'
        line += ";"
        if len(line) < self.column_width or len(params) == 0:
            self.inline(line, end='\n')
        else:
            self.write(f'{kind}(')
            with self.indent():
                if len(params) > 0:
                    self.inline(params[0], indent=True)
                    for i in range(1, len(params)):
                        self.inline(',', end='\n')
                        self.inline(params[i], indent=True)
                    self.inline(end='\n')
                else:
                    self.write('// empty')
            self.inline(')', indent=True)
            if specs is not None:
                self.inline(f' {specs}')
            self.inline(';', end='\n')

    def gen_copy_constructor_decl(
            self,
            kind: str,
            specs: Optional[str] = None
    ) -> None:
        self.gen_constructor_decl(
            kind,
            params=[
                f'const {kind} &other',
            ],
            specs=specs
        )

    def gen_move_constructor_decl(
            self,
            kind: str,
            specs: Optional[str] = None
    ) -> None:
        self.gen_constructor_decl(
            kind,
            params=[
                f'{kind} &&other',
            ],
            specs=specs
        )

    def get_sup_name_and_params(
            self,
            sup_spec_opt_params: Union[
                LspSpec,
                Tuple[
                    Union[LspSpec, str],
                    List[str]
                ]
            ]
    ) -> Tuple[str, str]:
        match sup_spec_opt_params:
            case dict():
                sup_spec = sup_spec_opt_params
                sup_name = sup_spec["name"]
                params = ""
            case tuple() | list():
                sup_spec_or_name, params = sup_spec_opt_params
                match sup_spec_or_name:
                    case dict():
                        sup_spec = sup_spec_or_name
                        sup_name = sup_spec["name"]
                    case str():
                        sup_name = sup_spec_or_name
                params = ", ".join(params)
        return sup_name, params

    @contextmanager
    def gen_constructor(
            self,
            kind: str,
            params: List[Union[str, LspSpec]] = [],
            sups: Optional[
                List[
                    Union[
                        LspSpec,
                        Tuple[
                            Union[LspSpec, str],
                            List[str]
                        ]
                    ]
                ]
            ] = None,
            inits: Optional[
                List[
                    Tuple[str, List[str]]
                ]
            ] = None,
            specs: Optional[str] = None,
    ) -> Iterator:
        params = [(self.get_type_declaration(param)
                   if isinstance(param, dict)
                   else param)
                  for param in params]
        line = f'{self.indentation()}{kind}::{kind}({", ".join(params)})'
        if specs is not None:
            if isinstance(specs, list):
                specs = " ".join(specs)
            line += f' {specs}'
        if len(line) < self.column_width or len(params) == 0:
            self.inline(line, end='\n')
        else:
            self.write(f'{kind}::{kind}(')
            with self.indent():
                if len(params) > 0:
                    self.inline(params[0], indent=True)
                    for i in range(1, len(params)):
                        self.inline(',', end='\n')
                        self.inline(params[i], indent=True)
                    self.inline(end='\n')
                else:
                    self.write('// empty')
            self.inline(')', indent=True)
            if specs is not None:
                self.inline(f' {specs}')
            self.inline(end='\n')
        if sups is not None:
            with self.indent():
                sup_name, params = self.get_sup_name_and_params(sups[0])
                self.write(f': {sup_name}({params})')
                for i in range(1, len(sups)):
                    sup_name, params = self.get_sup_name_and_params(sups[i])
                    self.write(f', {sup_name}({params})')
        if inits is not None:
            with self.indent():
                init_name, init_params = inits[0]
                if sups is not None:
                    self.write(f', {init_name}({", ".join(init_params)})')
                else:
                    self.write(f': {init_name}({", ".join(init_params)})')
                for i in range(1, len(inits)):
                    init_name, init_params = inits[i]
                    self.write(f', {init_name}({", ".join(init_params)})')
        with self.block():
            yield self

    def gen_destructor_decl(
            self,
            kind: str,
            params: List[Union[str, LspSpec]] = [],
            specs: Optional[str] = None
    ) -> None:
        params = [(self.get_type_declaration(param)
                   if isinstance(param, dict)
                   else param)
                  for param in params]
        line = f'{self.indentation()}~{kind}({", ".join(params)})'
        if specs is not None:
            if isinstance(specs, list):
                specs = " ".join(specs)
            line += f' {specs}'
        line += ";"
        if len(line) < self.column_width or len(params) == 0:
            self.inline(line, end='\n')
        else:
            self.write(f'~{kind}(')
            with self.indent():
                self.inline(params[0], indent=True)
                for i in range(1, len(params)):
                    self.inline(',', end='\n')
                    self.inline(params[i], indent=True)
                self.inline(end='\n')
            self.inline(')', indent=True)
            if specs is not None:
                self.inline(f' {specs}')
            self.inline(';', end='\n')

    @contextmanager
    def gen_destructor(
            self,
            kind: str,
            params: List[Union[str, LspSpec]] = [],
            specs: Optional[str] = None
    ) -> Iterator:
        params = [(self.get_type_declaration(param)
                   if isinstance(param, dict)
                   else param)
                  for param in params]
        line = f'{self.indentation()}{kind}::~{kind}({", ".join(params)})'
        if specs is not None:
            if isinstance(specs, list):
                specs = " ".join(specs)
            line += f' {specs}'
        if len(line) < self.column_width or len(params) == 0:
            self.inline(line, end='\n')
        else:
            self.write(f'{kind}::~{kind}(')
            with self.indent():
                self.inline(params[0], indent=True)
                for i in range(1, len(params)):
                    self.inline(',', end='\n')
                    self.inline(params[i], indent=True)
                self.inline(end='\n')
            self.inline(')', indent=True)
            if specs is not None:
                self.inline(f' {specs}')
            self.inline(end='\n')
        with self.block():
            yield self

    def gen_call(self, fn: str, *args: str) -> None:
        line = f'{self.indentation()}{fn}({", ".join(args)});'
        if len(line) < self.column_width or len(args) == 0:
            self.inline(line, end='\n')
        else:
            self.write(f'{fn}(')
            with self.indent():
                self.inline(args[0], indent=True)
                for i in range(1, len(args)):
                    self.inline(',', end='\n')
                    self.inline(args[i], indent=True)
                self.inline(end='\n')
            self.write(');')

    def gen_throw(self, exc: str, *args: str) -> None:
        self.gen_call(f'throw {exc}', *args)

    def resolve(self, name_or_spec_or_symbol: Union[str, LspSpec, LspSymbol]) -> LspSymbol:
        name: str
        symbol: Optional[LspSymbol]
        match name_or_spec_or_symbol:
            case str():
                name = name_or_spec_or_symbol
                symbol = self.pipeline.indexer.name_index.get(name, None)
            case dict():
                spec: LspSpec = name_or_spec_or_symbol
                match spec["kind"]:
                    case "base" | "reference":
                        name = spec["name"]
                    case "or" | "literal" | "array" | "map" | "tuple":
                        name = self.nested_name()
                    case _:
                        raise ValueError(f'Unsupported symbol type ({spec["kind"]}): {spec}')
                symbol = self.pipeline.indexer.name_index.get(name, None)
            case LspSymbol():
                symbol = name_or_spec_or_symbol
            case _:
                raise ValueError(
                    f'Unsupported symbol type ({type(name_or_spec_or_symbol)}): {name_or_spec_or_symbol}'
                )
        if symbol is not None and symbol.resolution is not None:
            return symbol.resolution
        raise ValueError(f'Failed to resolve symbol: {name_or_spec_or_symbol}')

    def generate_disclaimer(self) -> None:
        self.write('// ------------------------------------------------------------------------------')
        self.write('// NOTE: This file was generated from Microsoft\'s Language Server Protocol (LSP)')
        self.write('// specification. Please do not edit it by hand.')
        self.write('// ------------------------------------------------------------------------------')
        self.newline()

    def generate_docstring(self, docs: Optional[str]) -> None:
        if docs is not None:
            try:
                docs = docs.replace(r'​', '') \
                           .replace(r'*/', '*{@literal /}') \
                           .replace(r'/*', '{@literal /}*')
                self.write( '/**')
                for line in docs.split('\n'):
                    self.inline(' *', indent=True)
                    if len(line) > 0:
                        self.inline(f' {line}')
                    self.newline()
                self.write( ' */')
            except Exception as e:
                stack_trace = traceback.format_exc()
                self.write('/*')
                self.write(stack_trace)
                self.write('/* DEBUG: Recovered from error, here. */')
                print(f"Recovered from failure while generating documentation.", file=sys.stderr)
                print(stack_trace, file=sys.stderr)

    def get_union_enumeration(
        self,
        spec_or_name: Union[str, LspSpec]
    ) -> str:
        match spec_or_name:
            case str():
                name: str = spec_or_name
                return rename_enum(name)
            case dict():
                spec: LspSpec = spec_or_name
                match spec["kind"]:
                    case "base" | "reference":
                        return rename_enum(spec["name"])
                    case "stringLiteral":
                        return rename_enum("string")
                    case "integerLiteral":
                        return rename_enum("integer")
                    case "booleanLiteral":
                        return rename_enum("boolean")
                    case "array":
                        return f'{self.get_union_enumeration(spec["element"])}Array'
                    case "map":
                        key = self.get_union_enumeration(spec["key"])
                        val = self.get_union_enumeration(spec["value"])
                        return f'{key}{val}'
                    case "and":
                        raise ValueError(
                            f'AND types are not supported for type declarations: {spec}'
                        )
                    case "or":
                        buffer = []
                        for index, item_spec in enumerate(spec["items"]):
                            with self.nest_name(str(index)):
                                if index > 0:
                                    buffer.append('Or')
                                buffer.append(self.get_union_enumeration(item_spec))
                        return "".join(buffer)
                    case "tuple":
                        buffer = []
                        item_specs = spec["items"]
                        if len(item_specs) == 2:
                            buffer.append('PairOf')
                        else:
                            buffer.append('TupleOf')
                        for index, item_spec in enumerate(item_specs):
                            if index > 0:
                                buffer.append('And')
                            buffer.append(self.get_union_enumeration(item_spec))
                        return "".join(buffer)
                    case "literal":
                        return rename_enum(self.nested_name())
                    case _:
                        raise ValueError(f'Unsupported union enumeration kind ({spec["kind"]}): {spec}')

    def generate_union_enumeration(
        self,
        spec: LspSpec
    ) -> None:
        self.inline(self.get_union_enumeration(spec))

    def get_enum_value(self, enumeration: str, enumerator: Union[str, LspSpec]) -> str:
        match enumerator:
            case str():
                return f'{enumeration}::{enumerator}'
            case dict():
                enumerator = self.get_union_enumeration(enumerator)
                return self.get_enum_value(enumeration, enumerator)
            case _:
                raise ValueError(f'Unsupported enumerator type ({type(enumerator)}): {enumerator}')

    def name_field(self, type_spec: LspSpec) -> str:
        match type_spec["kind"]:
            case "base" | "reference":
                field_name = rename_field(type_spec["name"])
            case "literal":
                field_name = self.nested_name()
                field_name = rename_field(field_name)
            case "tuple":
                elem_specs = type_spec["items"]
                if len(elem_specs) == 2:
                    field_name = "pairOf"
                else:
                    field_name = "tupleOf"
                for index, elem_spec in enumerate(elem_specs):
                    if index > 0:
                        field_name += "And"
                    match elem_spec["kind"]:
                        case "base" | "reference":
                            field_name += upper_first(elem_spec["name"])
                        case _:
                            raise ValueError(f'Unsupported tuple field type ({elem_spec["kind"]}): {elem_spec}')
            case "array":
                elem_spec = type_spec["element"]
                match elem_spec["kind"]:
                    case "base" | "reference":
                        elem_type_name = elem_spec["name"]
                        elem_symbol = self.resolve(elem_type_name)
                        field_name = elem_symbol.normalized_name
                        field_name = f'{field_name}Array'
                        field_name = rename_field(field_name)
                    case "or":
                        field_name = self.nested_name()
                        field_name = rename_field(field_name)
                    case _:
                        raise ValueError(f'Unsupported array field type ({elem_spec["kind"]}): {elem_spec}')
            case _:
                raise ValueError(f'Unsupported field type ({type_spec["kind"]}): {type_spec}')
        return field_name

    @contextmanager
    def type_declaration_options(
            self,
            wrap_ptr: Optional[bool] = None,
            wrap_nested_ptr: Optional[bool] = None,
            string_ptr: Optional[bool] = None,
            nested_string_ptr: Optional[bool] = None
    ) -> Iterator:
        if wrap_ptr is None:
            wrap_ptr = self._type_declaration_wrap_ptr
        if wrap_nested_ptr is None:
            wrap_nested_ptr = self._type_declaration_wrap_nested_ptr
        if string_ptr is None:
            string_ptr = self._type_declaration_string_ptr
        if nested_string_ptr is None:
            nested_string_ptr = self._type_declaration_nested_string_ptr

        old_wrap_ptr = self._type_declaration_wrap_ptr
        old_wrap_nested_ptr = self._type_declaration_wrap_nested_ptr
        old_string_ptr = self._type_declaration_string_ptr
        old_nested_string_ptr = self._type_declaration_nested_string_ptr

        self._type_declaration_wrap_ptr = wrap_ptr
        self._type_declaration_wrap_nested_ptr = wrap_nested_ptr
        self._type_declaration_string_ptr = string_ptr
        self._type_declaration_nested_string_ptr = nested_string_ptr

        yield self

        self._type_declaration_wrap_ptr = old_wrap_ptr
        self._type_declaration_wrap_nested_ptr = old_wrap_nested_ptr
        self._type_declaration_string_ptr = old_string_ptr
        self._type_declaration_nested_string_ptr = old_nested_string_ptr

    def _wrap_ptr(self, level: int) -> bool:
        return self._type_declaration_wrap_ptr and level == 0 \
            or self._type_declaration_wrap_nested_ptr and level > 0

    def _string_ptr(self, level: int) -> bool:
        return self._type_declaration_string_ptr and level == 0 \
            or self._type_declaration_nested_string_ptr and level > 0

    def wrap_ptr(self, type_name: str, level: int) -> str:
        if self._wrap_ptr(level):
            type_name = f'std::unique_ptr<{type_name}>'
        return type_name

    def get_type_declaration(
            self,
            type_spec_or_name_or_symbol: Union[str, LspSpec, LspSymbol],
            level: int = 0) -> str:
        next_level = level + 1
        match type_spec_or_name_or_symbol:
            case dict():
                type_spec: LspSpec = type_spec_or_name_or_symbol
                match type_spec["kind"]:
                    case "base":
                        type_name = type_spec["name"]
                        return self.get_type_declaration(type_name, level=level)
                    case "reference":
                        type_name = type_spec["name"]
                        match type_name:
                            case "LSPAny" | "LSPObject" | "LSPArray":
                                return self.wrap_ptr(type_name, level)
                            case _:
                                symbol = self.resolve(type_name)
                                match symbol.kind:
                                    case LspSymbolKind.BASE:
                                        type_name = symbol.normalized_name
                                        return self.get_type_declaration(type_name, level=level)
                                    case LspSymbolKind.STRUCTURE | \
                                        LspSymbolKind.UNION | \
                                        LspSymbolKind.INNER | \
                                        LspSymbolKind.ARRAY | \
                                        LspSymbolKind.MAP | \
                                        LspSymbolKind.TUPLE:
                                        return self.wrap_ptr(type_name, level)
                                    case _:
                                        return type_name
                    case "array":
                        array_spec = type_spec
                        elem_spec = array_spec['element']
                        array_name = self.nested_name()
                        with self.type_declaration_options(
                                string_ptr=False,
                                wrap_ptr=self.has_cycle(array_name, elem_spec)
                        ):
                            with self.nest_name('elem'):
                                elem = self.get_type_declaration(elem_spec, level=0)
                        type_name = f'std::vector<{elem}>'
                        return self.wrap_ptr(type_name, level)
                    case "map":
                        map_spec = type_spec
                        key_spec = map_spec["key"]
                        value_spec = map_spec["value"]
                        map_name = self.nested_name()
                        with self.type_declaration_options(
                                string_ptr=False,
                                wrap_ptr=self.has_cycle(map_name, value_spec)
                        ):
                            with self.nest_name('key'):
                                key = self.get_type_declaration(key_spec, level=0)
                            with self.nest_name('value'):
                                val = self.get_type_declaration(value_spec, level=0)
                        type_name = f'std::map<{key}, {val}>'
                        return self.wrap_ptr(type_name, level)
                    case "and":
                        raise ValueError(
                            f'AND types are not supported for type declarations: {type_spec}'
                        )
                    case "or":
                        type_name = self.nested_name()
                        return self.wrap_ptr(type_name, level)
                    case "tuple":
                        with self.type_declaration_options(string_ptr=False):
                            item_specs: List[LspSpec] = type_spec["items"]
                            items = ", ".join((self.get_type_declaration(item_spec)
                                               for item_spec in item_specs))
                            items: Union[str, List[str]] = []
                            for item_index, item_spec in enumerate(item_specs):
                                with self.nest_name(str(item_index)):
                                    item = self.get_type_declaration(item_spec, level=next_level)
                                    items.append(item)
                            items = ", ".join(items)
                            if len(item_specs) == 2:
                                type_name = f'std::pair<{items}>'
                            else:
                                type_name = f'std::tuple<{items}>'
                            return self.wrap_ptr(type_name, level)
                    case "literal":
                        type_name = self.nested_name()
                        return self.wrap_ptr(type_name, level)
                    case "stringLiteral":
                        return rename_type('string')
                    case "integerLiteral":
                        return rename_type('integer')
                    case "booleanLiteral":
                        return rename_type('boolean')
                    case _:
                        raise ValueError(f'Unsupported Type kind: {type_spec}')
            case str():
                type_name: str = type_spec_or_name_or_symbol
                symbol = self.resolve(type_name)
                if symbol.normalized_name == "string" and self._string_ptr(level):
                    return f'std::unique_ptr<{symbol.resolved_name}>'
                return symbol.resolved_name
            case LspSymbol():
                symbol: LspSymbol = type_spec_or_name_or_symbol
                return self.get_type_declaration(symbol.spec)
            case _:
                raise ValueError(
                    f'Unsupported spec type ({type(type_spec_or_name)}): {type_spec_or_name}'
                )

    def is_union_pointer_type(self, field_spec: LspSpec) -> bool:
        match field_spec["kind"]:
            case "base":
                base_symbol = self.resolve(field_spec["name"])
                return base_symbol.normalized_name == "string"
            case "reference":
                symbol = self.resolve(field_spec["name"])
                match symbol.kind:
                    case LspSymbolKind.BASE:
                        return symbol.normalized_name == "string"
                    case LspSymbolKind.ENUMERATION:
                        return False
                    case LspSymbolKind.STRUCTURE \
                         | LspSymbolKind.ARRAY \
                         | LspSymbolKind.TUPLE \
                         | LspSymbolKind.MAP \
                         | LspSymbolKind.UNION:
                        return True
                    case _:
                        raise ValueError(
                            f'Unsupported field type ({symbol.kind}): {symbol.spec}'
                        )
            case "enumeration":
                return False
            case "array" | "map" | "tuple" | "or" | "literal":
                return True
            case _:
                raise ValueError(
                    f'Unsupported field type ({field_spec["kind"]}): {field_spec}'
                )

    def generate_type_declaration(self, type_spec: LspSpec) -> None:
        self.inline(self.get_type_declaration(type_spec))

    def rename_base_type(self, base_name: str) -> str:
        match base_name:
            case "boolean":
                return "bool"
            case "decimal":
                return "double"
            case "integer":
                return "int"
            case "uinteger":
                return "unsigned int"
            case "string" | "DocumentUri" | "RegExp" | "URI":
                return "std::string"
            case "null":
                return "std::nullptr_t"
            case _:
                raise ValueError(f'Unrecognized LSP base type name: {base_name}')

    def generate_enumeration(self, enum_spec: LspSpec) -> None:
        pass

    def generate_message_enums(self, enum_name: str, symbols: List[LspSymbol]) -> None:
        enumeration = []
        for symbol in symbols:
            message_method = symbol.name
            enumerator = method_to_camel_case(message_method)
            enumeration.append((enumerator, message_method))
        enum_spec = as_enumeration_spec(enum_name, enumeration)
        self.generate_enumeration(enum_spec)

    def generate_request_enums(self) -> None:
        self.generate_message_enums(
            'IncomingRequest',
            self.pipeline.indexer.request_index["clientToServer"]
        )
        self.generate_message_enums(
            'OutgoingRequest',
            self.pipeline.indexer.request_index["serverToClient"]
        )

    def generate_notification_enums(self) -> None:
        self.generate_message_enums(
            'IncomingNotification',
            self.pipeline.indexer.notification_index["clientToServer"]
        )
        self.generate_message_enums(
            'OutgoingNotification',
            self.pipeline.indexer.notification_index["serverToClient"]
        )

    @contextmanager
    def gen_if_debug_enabled(self, if_expr: Optional[str] = None) -> Iterator:
        self.inline('#ifdef DEBUG', end='\n')
        if if_expr is not None:
            with self.gen_if(if_expr):
                yield self
        else:
            yield self
        self.inline('#endif // DEBUG', end='\n')

    @contextmanager
    def gen_throw_lsp_exception(self, exception_type: str) -> Iterator:
        self.write('throw LSP_EXCEPTION(')
        with self.indent():
            self.write(f'{exception_type},')
            yield self
        self.write(');')

    def gen_throw_invalid_params(self, *fragments: str) -> None:
        with self.gen_throw_lsp_exception('ErrorCodes::InvalidParams'):
            for fragment in fragments:
                self.write(fragment)

    def gen_return(self, expr: str) -> None:
        self.write(f'return {expr};')

    def gen_break(self) -> None:
        self.write('break;')

    def has_cycle(
            self,
            source_or_name: Union[LspSymbol, str],
            name_or_spec_or_symbol: Union[str, LspSpec, LspSymbol]
    ) -> bool:
        source: LspSymbol
        match source_or_name:
            case LspSymbol():
                source = source_or_name
            case str():
                name: str = source_or_name
                try:
                    source = self.resolve(name)
                except ValueError:
                    return False
        if source.cycles is None:
            return False
        name: str
        spec: LspSpec
        target: LspSymbol
        match name_or_spec_or_symbol:
            case str():
                name = name_or_spec_or_symbol
                target = self.resolve(name)
            case dict():
                spec = name_or_spec_or_symbol
                match spec["kind"]:
                    case "base" | "reference":
                        name = spec["name"]
                    case "or" | "literal" | "array" | "map" | "tuple":
                        name = self.nested_name()
                    case "stringLiteral" | "integerLiteral" | "booleanLiteral":
                        return False
                    case _:
                        raise NotImplementedError(
                            f'Unsupported spec type ({spec["kind"]}): {spec}'
                        )
                target = self.resolve(name)
            case LspSymbol():
                target = name_or_spec_or_symbol
        return target in source.cycles



"""
Script for conveniently updating the translation files.

Run this script with:

.. code::

    python3 intl.py


You can also pass the language code as an argument:

.. code::

    python3 intl.py de


Do not pass the default language code as an argument.
"""

import json
import sys
import subprocess
from pathlib import Path
from typing import List

root = Path(__file__).parent
"""
Make sure to run this script from the root of the documenation repository.
"""

all_languages: List[str]
"""
List of currently supported languages, taken from ``doc/src/_static/languages.json``.

To support a new language add its `language code`_ to the list.

.. _language code: https://www.sphinx-doc.org/en/master/usage/configuration.html#intl-options
"""

with open(root / "src" / "_static" / "languages.json") as fd:
    all_languages = list(json.load(fd).values())


def intl_gettext() -> None:
    """
    Update the ``gettext`` translation files.
    """

    subprocess.run(
        ["sphinx-build", "-b", "gettext", str(root / "src"), str(root / "_gettext")],
        cwd=root,
    )


def intl_update(language: str) -> None:
    """
    Update a translation file.

    Parameters
    ----------
    language : str
        Language to update.
    """

    subprocess.run(
        [
            "sphinx-intl",
            "update",
            "-l",
            language,
            "-d",
            str(root / "locale"),
            "-p",
            str(root / "_gettext"),
        ],
        cwd=root,
    )


def intl_all(languages: List[str]) -> None:
    """
    Update all translation files.

    Parameters
    ----------
    languages : List[str]
        List of languages to update.
    """

    intl_gettext()
    intl_update(",".join(languages))


if __name__ == "__main__":
    languages = sys.argv[1:] if len(sys.argv) > 1 else all_languages[1:]

    intl_all(languages)

    print()
    print("Please commit the changes to the translation files if necessary.")
    print()
    print(f"    git add \"{root / 'locale'}\"")
    print()



"""
# Design

This file converts from a GFortran module file representation (documented in
the `lfortran.adapters.gfortran.mod` module) to an Abstract Semantic Representation (ASR).
"""

# TODO: move this into the lfortran package itself
import sys
sys.path.append("../..")

from lfortran.asr import asr
from lfortran.asr.asr_check import verify_asr
from lfortran.asr.asr_to_ast import asr_to_ast
from lfortran.ast.ast_to_src import ast_to_src
from lfortran.adapters.gfortran.mod import mod_to_asr
from lfortran.asr.builder import (make_translation_unit,
    translation_unit_make_module, scope_add_function, make_type_integer,
    make_type_real, type_eq, make_binop, scope_add_symbol, Scope,
    function_make_var, array_is_assumed_shape)

class NodeTransformer(asr.NodeTransformerBase):

    def visit_scope(self, symtab, parent=None):
        new_symtab = Scope(parent)
        self._lookup = 2
        self._scope = new_symtab
        for s, sym in symtab.symbols.items():
            new_symtab.symbols[s] = self.visit(sym)
        self._lookup = 0
        return new_symtab

    def visit_sequence(self, seq):
        r = []
        if seq is not None:
            for node in seq:
                r.append(self.visit(node))
        return r

    def visit_object(self, o):
        if isinstance(o, Scope):
            return self.visit_scope(o)
        elif isinstance(o, list):
            return self.visit_sequence(o)
        elif isinstance(o, (str, int)) or o is None:
            return o
        else:
            print(type(o))
            raise NotImplementedError()

    def visit_Function(self, node):
        name = self.visit_object(node.name)
        symtab = self.visit_object(node.symtab)
        self._lookup = 1
        self._scope = symtab
        args = self.visit_sequence(node.args)
        body = self.visit_sequence(node.body)
        return_var = self.visit(node.return_var)
        self._lookup = 0
        if node.bind:
            bind = self.visit(node.bind)
        else:
            bind = None
        return asr.Function(name=name, args=args, body=body, bind=bind, return_var=return_var, symtab=symtab)

    def visit_VariableOld(self, node):
        if self._lookup == 1:
            return self._scope.resolve(node.name)
        elif self._lookup == 2:
            v = self._scope.resolve(node.name, raise_exception=False)
            if v:
                return v
        name = self.visit_object(node.name)
        intent = self.visit_object(node.intent)
        dummy = self.visit_object(node.dummy)
        type = self.visit(node.type)
        return asr.VariableOld(name=name, intent=intent, dummy=dummy, type=type)


class WrapperVisitor(NodeTransformer):

    def visit_Module(self, node):
        name = "mod2" #node.name + "_wrapper"
        self._modname = node.name
        symtab = self.visit_scope(node.symtab)
        return asr.Module(name=name, symtab=symtab)

    def visit_Function(self, node):
        name = self.visit_object(node.name)
        symtab = self.visit_object(node.symtab)
        self._lookup = 1
        self._scope = symtab
        args = self.visit_sequence(node.args)
        body = self.visit_sequence(node.body)
        self._lookup = 0
        if node.bind:
            bind = self.visit(node.bind)
        else:
            bind = None
        tmp = asr.VariableOld(name="a", type=make_type_integer())
        f = asr.Function(name=name, args=args, body=body, bind=bind,
            symtab=symtab, return_var=tmp)
        return_var = function_make_var(f, name="r", type=self.visit(node.return_var.type))
        return_var.dummy = True
        f.return_var = return_var

        cname = node.name + "_c_wrapper"
        mangled_name = '__' + self._modname + '_MOD_' + node.name.lower()
        bind = asr.Bind(lang="c", name=mangled_name)
        cargs = []
        args2 = []
        type1 = make_type_integer()
        type1.dims = [asr.dimension(asr.Num(n=1, type=make_type_integer()))]
        type2 = make_type_integer()
        type2.dims = [
            asr.dimension(asr.Num(n=1, type=make_type_integer())),
            asr.dimension(asr.Num(n=1, type=make_type_integer()))
            ]
        c_desc1 = scope_add_function(
            symtab,
            "c_desc1_int32",
            args=[
                asr.VariableOld(
                    name="A",
                    intent="in",
                    type=type1
                )
            ],
            return_var=asr.VariableOld(
                name="c_desc1_int32",
                type=asr.Derived(name="c_desc1_t")
            ),
            module = "gfort_interop"
        )
        c_desc2 = scope_add_function(
            symtab,
            "c_desc2_int32",
            args=[
                asr.VariableOld(
                    name="A",
                    intent="in",
                    type=type2
                )
            ],
            return_var=asr.VariableOld(
                name="c_desc2_int32",
                type=asr.Derived(name="c_desc2_t")
            ),
            module = "gfort_interop"
        )
        for arg in node.args:
            type = self.visit(arg.type)
            if array_is_assumed_shape(type):
                if len(type.dims) == 1:
                    dname = "c_desc1_t"
                    fname = c_desc1
                elif len(type.dims) == 2:
                    dname = "c_desc2_t"
                    fname = c_desc2
                else:
                    raise NotImplementedError("Too many dimensions")
                type = asr.Derived(name=dname, module="gfort_interop")
                args2.append(asr.FuncCall(func=fname, args=[arg],
                    keywords=[], type=type))
            else:
                args2.append(arg)
            cargs.append(asr.VariableOld(
                name=arg.name,
                intent=arg.intent,
                type=type,
            ))
        fw = scope_add_function(symtab, cname, args=cargs, return_var=cname)
        fw.bind = bind

        body = [
            asr.Assignment(return_var,
                asr.FuncCall(func=fw, args=args2, keywords=[],
                    type=fw.return_var.type)
            )
        ]
        f.body = body

        return f

def create_wrapper(u):
    v = WrapperVisitor()
    u2 = v.visit(u)
    verify_asr(u2)
    return u2

u = mod_to_asr("mod1.mod")
a = asr_to_ast(u)
s = ast_to_src(a)
#print(s)

u2 = create_wrapper(u)
a = asr_to_ast(u2)
s = ast_to_src(a)
#print("-"*80)
print(s)



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Set

from llanguage_server.lsp.utils import normalize_name, rename_type

LspSpec = Dict[str, Any]


class LspSymbolKind(str, Enum):
    ALIAS = 'alias'
    ARRAY = 'array'
    BASE = 'base'
    ENUMERATION = 'enumeration'
    INNER = 'inner'
    MAP = 'map'
    NOTIFICATION = 'notification'
    REFERENCE = 'reference'
    REQUEST = 'request'
    STRUCTURE = 'structure'
    TUPLE = 'tuple'
    UNION = 'union'


@dataclass
class LspSymbol:
    name: str
    kind: LspSymbolKind
    spec: LspSpec
    dependencies: Optional[List["LspSymbol"]] = None
    fields: Optional[List[LspSpec]] = None
    resolution: Optional["LspSymbol"] = None
    normalized_name: str = field(init=False)
    resolved_name: str = field(init=False)
    cycles: Optional[Set["LspSymbol"]] = None

    def __post_init__(self) -> None:
        self.normalized_name = normalize_name(self.name)
        self.resolved_name = rename_type(self.name)

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, LspSymbol):
            return self.name == other.name
        return False



#!/usr/bin/env python3

"""
This script is used to convert the output of dwarfdump into a file that is easy
to load into lfortran to lookup filenames and line numbers for a given address.
Here is how to use it:

        cd src/bin
        llvm-dwarfdump --debug-line lfortran.dSYM > lfortran.dSYM/symbols.txt
        ./dwarf_convert.py lfortran.dSYM/symbols.txt lfortran.dSYM/lines.txt lfortran.dSYM/lines.dat

This is meant to be executed at build time.

A better solution would be to use the `dwarf` library directly from C++ and
generate the same output directly. Here is the source code of llvm-dwarfdump:

https://github.com/llvm/llvm-project/blob/91a6ad5ad887a16e361338303d4ff3d29dba5e10/llvm/tools/llvm-dwarfdump/llvm-dwarfdump.cpp

We just have to do exactly what it does, but generate the output in the format
of lines.txt and lines.dat

"""

from collections import namedtuple
from glob import glob
import os
from pathlib import Path
import re
from struct import pack
import sys

DebugLines = namedtuple("DebugLines", ["lines"])
DebugLine = namedtuple("DebugLine", ["include_directories", "file_names", "addresses"])
IncludeDirectory = namedtuple("IncludeDirectory", ["id", "path"])
FileName = namedtuple("FileName", ["id", "filename", "dir_idx"])

ASRDebugLines = namedtuple("ASRDebugLines", ["filenames", "addresses"])

def normalize_to_absolute_path(path_str):
    if path_str[0] == "/":
        # Absolute path
        full_path = Path(path_str)
    else:
        # Path is relative to the root directory of the repository
        # Path.cwd() is `src/bin`.
        base_dir = (Path.cwd() / ".." / "..").resolve()
        full_path = base_dir / path_str
    full_path = full_path.resolve()
    return str(full_path)

class Parser:
    """
    Parser for the output generated by dwarfdump.

    On macOS (both Intel and ARM based):

        dwarfdump --debug-line src/bin/lfortran.dSYM > symbols.txt

    Then parse it using:

        p = Parser()
        ast = p.parse_file("symbols.txt")
    """

    def parse_file(self, filename):
        self.file = open(filename)
        self.line = self.file.readline()
        while not self.line.startswith(".debug_line contents:"):
            self.line = self.file.readline()

        self.line = self.file.readline()
        lines = []
        while self.line.startswith("debug_line"):
            d = self.parse_debug_line()
            lines.append(d)

        if (self.line.rstrip() != ""):
            raise ValueError(f"Failed to completely parse {filename} !!")
        return DebugLines(lines)

    def parse_debug_line(self):
        self.line = self.file.readline()
        include_dirs_found = True
        while not self.line.startswith("include_directories"):
            if self.line.startswith("file_names"):
                include_dirs_found = False
                break
            else:
                self.line = self.file.readline()

        include_directories = []
        if include_dirs_found:
            while self.line.startswith("include_directories"):
                n, path = re.compile(r"include_directories\[[ ]*(\d+)\] = \"([^\"]+)\"").findall(self.line)[0]
                n = int(n)
                include_directories.append(IncludeDirectory(n, path))
                self.line = self.file.readline()

        file_names = []
        while self.line.startswith("file_names"):
            n = re.compile(r"file_names\[[ ]*(\d+)\]:").findall(self.line)[0]
            n = int(n)

            # these are the only two more values we currently need to make
            # 'FileName'
            filename = None
            dir_idx = None

            while True:
                self.line = self.file.readline()
                if not self.line or self.line.startswith("file_names") or not self.line.strip():
                    break

                name_match = re.compile(r"name: \"([^\"]+)\"").findall(self.line)
                if name_match:
                    filename = name_match[0]
                    continue

                dir_match = re.compile(r"dir_index: (\d+)").findall(self.line)
                if dir_match:
                    dir_idx = int(dir_match[0])
                    continue

                # this appear in newer version of dsymutil (Apple clang 17 or higher),
                # currently we ignore this
                md5_match = re.compile(r"md5_checksum: ([0-9a-fA-F]+)").findall(self.line)
                if md5_match:
                    continue

                # for older version of dsymutil `mod_time` and `length` are ignore,
                # kindly update this comment when we notice any newer keyword is noticed
                # in newer version(s) of dsymutil
                continue

            if filename is None or dir_idx is None:
                raise ValueError(f"Missing entries in construction of `FileName`")

            file_names.append(FileName(n, filename, dir_idx))


        self.line = self.file.readline()
        self.line = self.file.readline()
        self.line = self.file.readline()

        addresses = []
        while self.line.startswith("0x"):
            address, line, column, file_id = self.line.split()[:4]
            address = int(address, base=16)
            line = int(line)
            column = int(column)
            file_id = int(file_id)
            addresses.append([address, line, column, file_id])
            self.line = self.file.readline()

        self.line = self.file.readline()

        d = DebugLine(include_directories, file_names, addresses)
        return d

def ast_to_asr(ast):
    lines = []
    last_address = -1
    global_filename_id = 0
    global_filenames = []
    global_addresses = []
    for line in ast.lines:
        include_dirs = {}
        for inc in line.include_directories:
            include_dirs[inc.id] = inc.path
        filenames = {}
        for filename in line.file_names:
            prefix = ""
            if filename.dir_idx != 0:
                prefix = include_dirs[filename.dir_idx] + "/"
            filenames[filename.id] = global_filename_id
            global_filenames.append(normalize_to_absolute_path(prefix+filename.filename))
            global_filename_id += 1
        for address, line_num, column, file_id in line.addresses:
            filename = global_filenames[filenames[file_id]]
            assert last_address <= address
            last_address = address
            if line_num != 0:
                global_addresses.append([address, line_num, filenames[file_id]])
    return ASRDebugLines(global_filenames, global_addresses)


p = Parser()
ast = p.parse_file(sys.argv[1])
asr = ast_to_asr(ast)
with open(sys.argv[2], "w") as f:
    f.write(str(len(asr.filenames)) + "\n")
    for filename in asr.filenames:
        f.write(filename + "\n")
    f.write(str(len(asr.addresses)) + "\n")
with open(sys.argv[3], "wb") as f:
    for addr, line, fileid in asr.addresses:
        f.write(pack("3Q", addr, line, fileid))



import sys
import os
import re

HEAD = r"""#ifndef LFORTRAN_%(MOD)s_H
#define LFORTRAN_%(MOD)s_H

// Generated by grammar/wasm_instructions_visitor.py

#include <libasr/alloc.h>
#include <libasr/location.h>
#include <libasr/colors.h>
#include <libasr/containers.h>
#include <libasr/exception.h>
#include <libasr/asr_scopes.h>
#include <libasr/codegen/wasm_utils.h>


namespace LCompilers {

namespace %(MOD)s {

"""

FOOT = r"""
} // namespace %(MOD)s

} // namespace LCompilers

#endif // LFORTRAN_%(MOD)s_H
"""

class WASMInstructionsVisitor():
    def __init__(self, stream, data):
        self.stream = stream
        self.data = data

    def visit_BaseWASMVisitor(self, mod, *args):
        self.emit("template <class StructType>", 0)
        self.emit("class BaseWASMVisitor {", 0)
        self.emit("private:", 0)
        self.emit(    "StructType& self() { return static_cast<StructType&>(*this); }", 1)
        self.emit("public:", 0)
        self.emit(    "Vec<uint8_t> &code;", 1)
        self.emit(    "uint32_t offset;\n", 1)

        self.emit(    "BaseWASMVisitor(Vec<uint8_t> &code, uint32_t offset): code(code), offset(offset) {}", 1)

        for inst in mod["instructions"]:
            self.emit("void visit_%s(%s) {throw LCompilersException(\"visit_%s() not implemented\");}\n" % (inst["func_read"], make_param_list(inst["params"], comment_out=True), inst["func_read"]), 1)

        self.emit(    "void decode_instructions() {", 1)
        self.emit(        "uint8_t cur_byte = wasm::read_b8(code, offset);", 2)
        self.emit(        "while (cur_byte != 0x0B) {", 2)
        self.emit(            "switch (cur_byte) {", 3)
        for inst in filter(lambda i: i["opcode"] not in ["0xFC", "0xFD"], mod["instructions"]):
            self.emit(            "case %s: {" % (inst["opcode"]), 4)
            for param in inst["params"]:
                self.emit(            "%s %s = %s(code, offset);" % (param["type"], param["name"], param["read_func"]), 5)
            self.emit(                "self().visit_%s(%s);" % (inst["func_read"], make_param_list(inst["params"], call=True)), 5)
            self.emit(                "break;", 5)
            self.emit(            "}", 4)

        self.emit(                "case 0xFC: {", 4)
        self.emit(                    "uint32_t num = wasm::read_u32(code, offset);", 5)
        self.emit(                    "switch(num) {", 5)
        for inst in filter(lambda i: i["opcode"] == "0xFC", mod["instructions"]):
            self.emit(                    "case %sU: {" % (inst["params"][0]["val"]), 6)
            for param in inst["params"][1:]: # first param is already read right at the start of case 0xFC
                self.emit(                    "%s %s = %s(code, offset);" % (param["type"], param["name"], param["read_func"]), 7)
            self.emit(                        "self().visit_%s(%s);" % (inst["func_read"], make_param_list(inst["params"], call=True)), 7)
            self.emit(                        "break;", 7)
            self.emit(                    "}", 6)
        self.emit(                        "default: {", 6)
        self.emit(                            "throw LCompilersException(\"Unknown num for opcode 0xFC\");", 7)
        self.emit(                        "}", 6)
        self.emit(                    "}", 5)
        self.emit(                    "break;", 5)
        self.emit(                "}", 4)

        self.emit(                "case 0xFD: {", 4)
        self.emit(                    "uint32_t num = wasm::read_u32(code, offset);", 5)
        self.emit(                    "switch(num) {", 5)
        for inst in filter(lambda i: i["opcode"] == "0xFD", mod["instructions"]):
            self.emit(                    "case %sU: {" % (inst["params"][0]["val"]), 6)
            for param in inst["params"][1:]:  # first param is already read right at the start of case 0xFD
                self.emit(                    "%s %s = %s(code, offset);" % (param["type"], param["name"], param["read_func"]), 7)
            self.emit(                        "self().visit_%s(%s);" % (inst["func_read"], make_param_list(inst["params"], call=True)), 7)
            self.emit(                        "break;", 7)
            self.emit(                    "}", 6)
        self.emit(                        "default: {", 6)
        self.emit(                            "throw LCompilersException(\"Unknown num for opcode 0xFD\");", 7)
        self.emit(                        "}", 6)
        self.emit(                    "}", 5)
        self.emit(                    "break;", 5)
        self.emit(                "}", 4)

        self.emit(                "default: {", 4)
        self.emit(                    "throw LCompilersException(\"Unknown opcode\");", 5)
        self.emit(                "}", 4)
        self.emit(            "}", 3)
        self.emit(            "cur_byte = wasm::read_b8(code, offset);", 3)
        self.emit(        "}", 2)
        self.emit(    "}", 1)
        self.emit("};\n", 0)

    def visit_WASMInstsAssembler(self, mod):
        self.emit("template <class StructType>", 0)
        self.emit("class WASMInstsAssembler {", 0)
        self.emit("private:", 0)
        self.emit(    "Allocator &m_al;", 1)
        self.emit(    "Vec<uint8_t> &m_code;\n", 1)
        self.emit(    "StructType &self() { return static_cast<StructType &>(*this); }", 1)
        self.emit("public:", 0)
        self.emit(    "WASMInstsAssembler(Allocator &al, Vec<uint8_t> &code): m_al(al), m_code(code) {}\n", 1)

        for inst in filter(lambda i: i["opcode"] not in ["0xFC", "0xFD"], mod["instructions"]):
            self.emit("void emit_%s(%s) {" % (inst["func_emit"], make_param_list(inst["params"])), 1)
            self.emit(    "m_code.push_back(m_al, %s);" % (inst["opcode"]), 2)
            for param in inst["params"]:
                self.emit("%s(m_code, m_al, %s);" % (param["emit_func"], param["name"]), 2)
            self.emit("}\n", 1)

        for inst in filter(lambda i: i["opcode"] == "0xFC", mod["instructions"]):
            self.emit("void emit_%s(%s) {" % (inst["func_emit"], make_param_list(inst["params"])), 1)
            self.emit(    "m_code.push_back(m_al, %s);" % (inst["opcode"]), 2)
            self.emit(    "wasm::emit_u32(m_code, m_al, %s);" % (inst["params"][0]["val"]), 2)
            for param in inst["params"][1:]:
                self.emit("%s(m_code, m_al, %s);" % (param["emit_func"], param["name"]), 2)
            self.emit("}\n", 1)

        for inst in filter(lambda i: i["opcode"] == "0xFD", mod["instructions"]):
            self.emit("void emit_%s(%s) {" % (inst["func_emit"], make_param_list(inst["params"])), 1)
            self.emit(    "m_code.push_back(m_al, %s);" % (inst["opcode"]), 2)
            self.emit(    "wasm::emit_u32(m_code, m_al, %s);" % (inst["params"][0]["val"]), 2)
            for param in inst["params"][1:]:
                self.emit("%s(m_code, m_al, %s);" % (param["emit_func"], param["name"]), 2)
            self.emit("}\n", 1)

        self.emit("};\n", 0)

    def emit(self, line, level=0):
        indent = "    "*level
        self.stream.write(indent + line + "\n")

def filter_params(params):
    return list(filter(lambda param: param["val"] != "0x00" and param["name"] != "num", params))

def make_param_list(params, comment_out = False, call = False):
    params = filter_params(params)
    if call:
        return ", ".join(map(lambda p: p["name"], params))
    if comment_out:
        return ", ".join(map(lambda p: p["type"] + " /*" + p["name"] + "*/", params))
    return ", ".join(map(lambda p: p["type"] + " " + p["name"], params))


def read_file(path):
    with open(path, encoding="utf-8") as fp:
        return fp.read()

def process_raw_instructions(instructions_raw):
    instructions = []
    for line in instructions_raw.splitlines():
        if line.startswith("--"):
            continue
        if len(instructions) > 0 and line.startswith(" "):
            instructions[-1].append(line.strip())
        else:
            instructions.append(line.strip())
    return instructions

def get_func_name(func, emit = False):
    splitted_name = re.split("[._]", func)
    if emit:
        return "_".join(splitted_name)
    return "".join(map(lambda name_sub_part: name_sub_part.capitalize(), splitted_name))

param_read_function = {
    "uint8_t": "wasm::read_b8",
    "uint32_t": "wasm::read_u32",
    "int32_t": "wasm::read_i32",
    "int64_t": "wasm::read_i64",
    "float": "wasm::read_f32",
    "double": "wasm::read_f64"
}

param_emit_function = {
    "uint8_t": "wasm::emit_b8",
    "uint32_t": "wasm::emit_u32",
    "int32_t": "wasm::emit_i32",
    "int64_t": "wasm::emit_i64",
    "float": "wasm::emit_f32",
    "double": "wasm::emit_f64"
}

param_type = {
    "u8": "uint8_t",
    "u32": "uint32_t",
    "i32": "int32_t",
    "i64": "int64_t",
    "f32": "float",
    "f64": "double",
}

def parse_param_info(param_info):
    type_info, name, val = param_info.split(":")
    type = param_type[type_info]
    read_func = param_read_function[type]
    emit_func = param_emit_function[type]
    return {"type": type, "read_func": read_func, "emit_func": emit_func,
            "name": name, "val": val}

def parse_instructions(instructions):
    instructions_info = []
    for inst in instructions:
        binary_info, text_info = inst.split(" ⇒ ")
        binary_info = binary_info.strip().split(" ")
        opcode = binary_info[0]
        params_info = binary_info[1:]
        text_info = text_info.strip().split(" ")
        func_read = get_func_name(text_info[0]) # first parameter is the function name
        func_emit = get_func_name(text_info[0], emit=True) # first parameter is the function name
        text_params = text_info[1:] # text_params are currently not needed and hence not used further
        params = list(map(lambda param_info: parse_param_info(param_info), params_info))
        instructions_info.append({"opcode": opcode, "func_read": func_read,
                                   "func_emit": func_emit, "params": params})
    return instructions_info

def main(argv):
    if len(argv) == 3:
        def_file, out_file = argv[1:]
    elif len(argv) == 1:
        print("Assuming default values of wasm_instructions.txt and wasm_visitor.h")
        here = os.path.dirname(__file__)
        def_file = os.path.join(here, "wasm_instructions.txt")
        out_file = os.path.join(here, "wasm_visitor.h")
    else:
        print("invalid arguments")
        return 2
    instructions_raw = read_file(def_file)
    instructions = process_raw_instructions(instructions_raw)
    instructions_info = parse_instructions(instructions)
    subs = {"MOD": "WASM_INSTS_VISITOR"}
    fp = open(out_file, "w", encoding="utf-8")
    try:
        fp.write(HEAD % subs)
        wasm_instructions_visitor = WASMInstructionsVisitor(fp, None)
        wasm_instructions_visitor.visit_BaseWASMVisitor({"instructions": instructions_info})
        wasm_instructions_visitor.visit_WASMInstsAssembler({"instructions": instructions_info})
        fp.write(FOOT % subs)
    finally:
        fp.close()

if __name__ == "__main__":
    sys.exit(main(sys.argv))
